"""Ontology domain."""

from enum import Enum
from typing import NewType

JobName = NewType("JobName", str)


class JobMLTask(str, Enum):
    """List of ML Tasks."""

    ASSET_ANNOTATION = "ASSET_ANNOTATION"
    CLASSIFICATION = "CLASSIFICATION"
    NAMED_ENTITIES_RECOGNITION = "NAMED_ENTITIES_RECOGNITION"
    NAMED_ENTITIES_RELATION = "NAMED_ENTITIES_RELATION"
    OBJECT_RELATION = "OBJECT_RELATION"
    OBJECT_DETECTION = "OBJECT_DETECTION"
    PAGE_LEVEL_CLASSIFICATION = "PAGE_LEVEL_CLASSIFICATION"
    PAGE_LEVEL_TRANSCRIPTION = "PAGE_LEVEL_TRANSCRIPTION"
    POSE_ESTIMATION = "POSE_ESTIMATION"
    RANKING = "RANKING"
    SPEECH_TO_TEXT = "SPEECH_TO_TEXT"
    TRANSCRIPTION = "TRANSCRIPTION"


class JobTool(str, Enum):
    """List of tools."""

    MARKER = "marker"
    POLYGON = "polygon"
    POLYLINE = "polyline"
    POSE = "pose"
    RANGE = "range"
    RECTANGLE = "rectangle"
    SEMANTIC = "semantic"
    VECTOR = "vector"



"""Notification use cases."""

from typing import Dict, Generator

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.domain.notification import NotificationFilter
from kili.domain.types import ListOrTuple
from kili.use_cases.base import BaseUseCases


class NotificationUseCases(BaseUseCases):
    """Notification use cases."""

    def list_notifications(
        self, filters: NotificationFilter, fields: ListOrTuple[str], options: QueryOptions
    ) -> Generator[Dict, None, None]:
        """List notifications."""
        return self._kili_api_gateway.list_notifications(
            filters=filters, fields=fields, options=options
        )

    def count_notifications(self, filters: NotificationFilter) -> int:
        """Count notifications."""
        return self._kili_api_gateway.count_notification(filters=filters)



"""Tools for export with images."""

from pathlib import Path
from typing import Dict, Tuple

from PIL import Image

from kili.services.export.exceptions import NotExportableAssetError


def _get_image_dimensions(filepath: str) -> Tuple:
    """Get an image width and height."""
    image = Image.open(filepath)
    return image.size


def get_image_dimensions(asset: Dict) -> Tuple:
    """Get an asset image width and height."""
    if "resolution" in asset and asset["resolution"] is not None:
        return (asset["resolution"]["width"], asset["resolution"]["height"])

    if Path(asset["content"]).is_file():
        return _get_image_dimensions(asset["content"])

    raise NotExportableAssetError(
        f"Could not find dimensions for asset with externalId '{asset['externalId']}'. Please"
        " use `kili.update_properties_in_assets()` to update the resolution of your asset. Or use"
        " `kili.export_labels(with_assets=True).`"
    )


def get_frame_dimensions(asset: Dict) -> Tuple:
    """Get a video asset frame width and height."""
    if "resolution" in asset and asset["resolution"] is not None:
        return (asset["resolution"]["width"], asset["resolution"]["height"])

    if (
        isinstance(asset["jsonContent"], list)
        and len(asset["jsonContent"]) > 0
        and Path(asset["jsonContent"][0]).is_file()
    ):
        return _get_image_dimensions(asset["jsonContent"][0])

    raise NotExportableAssetError(
        f"Could not find dimensions for asset with externalId '{asset['externalId']}'. Please"
        " use `kili.update_properties_in_assets()` to update the resolution of your asset. Or use"
        " `kili.export_labels(with_assets=True).`"
    )



"""Project command of Kili CLI."""

import click

from kili.entrypoints.cli.common_args import CONTEXT_SETTINGS
from kili.entrypoints.cli.project.copy import copy_project
from kili.entrypoints.cli.project.create import create_project
from kili.entrypoints.cli.project.describe import describe_project
from kili.entrypoints.cli.project.export import export_labels
from kili.entrypoints.cli.project.import_ import import_assets
from kili.entrypoints.cli.project.label import import_labels
from kili.entrypoints.cli.project.list_ import list_projects
from kili.entrypoints.cli.project.member import member


@click.group(context_settings=CONTEXT_SETTINGS)
def project():
    """Commands to interact with a Kili project."""


project.add_command(create_project, name="create")
project.add_command(describe_project, name="describe")
project.add_command(import_assets, name="import")
project.add_command(import_labels, name="label")
project.add_command(list_projects, name="list")
project.add_command(member, name="member")
project.add_command(export_labels, name="export")
project.add_command(copy_project, name="copy")



"""Decorators for label response module."""

from typing import Callable, Tuple


def for_all_properties(decorator: Callable, exclude: Tuple[str, ...] = ()) -> Callable:
    """Class decorator to decorate all the properties of the decorated class.

    Decorates with a decorator passed as argument.

    Args:
        decorator: Decorator to apply to all the properties of the decorated class.
        exclude: Sequence of properties to exclude from the decoration.
    """

    def decorate(cls):
        for attr in dir(cls):
            if attr in exclude:
                continue
            if isinstance(getattr(cls, attr), property):
                setattr(
                    cls,
                    attr,
                    property(
                        decorator(getattr(cls, attr).fget),
                        decorator(getattr(cls, attr).fset),
                        decorator(getattr(cls, attr).fdel),
                        getattr(cls, attr).__doc__,
                    ),
                )
        return cls

    return decorate



"""Mixin extending Kili API Gateway class with User related operations."""

from typing import Dict, Generator

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.domain.types import ListOrTuple
from kili.domain.user import UserFilter

from .mappers import create_user_data_mapper, update_user_data_mapper, user_where_mapper
from .operations import (
    GQL_COUNT_USERS,
    get_create_user_mutation,
    get_current_user_query,
    get_update_password_mutation,
    get_update_user_mutation,
    get_users_query,
)
from .types import CreateUserDataKiliGatewayInput, UserDataKiliGatewayInput


class UserOperationMixin(BaseOperationMixin):
    """GraphQL Mixin extending GraphQL Gateway class with User related operations."""

    def list_users(
        self, user_filters: UserFilter, fields: ListOrTuple[str], options: QueryOptions
    ) -> Generator[Dict, None, None]:
        """Return a generator of users that match the filter."""
        fragment = fragment_builder(fields)
        query = get_users_query(fragment)
        where = user_where_mapper(filters=user_filters)
        return PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving users", GQL_COUNT_USERS
        )

    def count_users(self, user_filters: UserFilter) -> int:
        """Return the number of users that match the filter."""
        variables = {"where": user_where_mapper(user_filters)}
        result = self.graphql_client.execute(GQL_COUNT_USERS, variables)
        return result["data"]

    def get_current_user(self, fields: ListOrTuple[str]) -> Dict:
        """Return the current user."""
        fragment = fragment_builder(fields)
        query = get_current_user_query(fragment=fragment)
        result = self.graphql_client.execute(query)
        return result["data"]

    def create_user(self, data: CreateUserDataKiliGatewayInput, fields: ListOrTuple[str]) -> Dict:
        """Create a user."""
        fragment = fragment_builder(fields)
        query = get_create_user_mutation(fragment)
        variables = {"data": create_user_data_mapper(data)}
        result = self.graphql_client.execute(query, variables)
        return result["data"]

    def update_password(
        self,
        old_password: str,
        new_password_1: str,
        new_password_2: str,
        user_filter: UserFilter,
        fields: ListOrTuple[str],
    ) -> Dict:
        """Update user password."""
        fragment = fragment_builder(fields)
        query = get_update_password_mutation(fragment)
        variables = {
            "data": {
                "oldPassword": old_password,
                "newPassword1": new_password_1,
                "newPassword2": new_password_2,
            },
            "where": user_where_mapper(filters=user_filter),
        }
        result = self.graphql_client.execute(query, variables)
        return result["data"]

    def update_user(
        self, user_filter: UserFilter, data: UserDataKiliGatewayInput, fields: ListOrTuple[str]
    ) -> Dict:
        """Update a user."""
        fragment = fragment_builder(fields)
        query = get_update_user_mutation(fragment)
        variables = {
            "data": update_user_data_mapper(data),
            "where": user_where_mapper(filters=user_filter),
        }
        result = self.graphql_client.execute(query, variables)
        return result["data"]



"""GraphQL payload data mappers for project operations."""

from typing import Dict

from kili.domain.project import ProjectFilters

from .types import ProjectDataKiliAPIGatewayInput


def project_where_mapper(filters: ProjectFilters) -> Dict:
    """Build the GraphQL ProjectWhere variable to be sent in an operation."""
    ret = {
        "id": filters.id,
        "searchQuery": filters.search_query,
        "shouldRelaunchKpiComputation": filters.should_relaunch_kpi_computation,
        "starred": filters.starred,
        "updatedAtGte": filters.updated_at_gte,
        "updatedAtLte": filters.updated_at_lte,
        "createdAtGte": filters.created_at_gte,
        "createdAtLte": filters.created_at_lte,
        "tagIds": filters.tag_ids,
    }
    if filters.archived is not None:
        ret["archived"] = filters.archived
    return ret


def project_data_mapper(data: ProjectDataKiliAPIGatewayInput) -> Dict:
    """Build the GraphQL ProjectData variable to be sent in an operation."""
    return {
        "archived": data.archived,
        "author": data.author,
        "complianceTags": data.compliance_tags,
        "consensusMark": data.consensus_mark,
        "consensusTotCoverage": data.consensus_tot_coverage,
        "description": data.description,
        "canNavigateBetweenAssets": data.can_navigate_between_assets,
        "canSkipAsset": data.can_skip_asset,
        "honeypotMark": data.honeypot_mark,
        "inputType": data.input_type,
        "instructions": data.instructions,
        "jsonInterface": data.json_interface,
        "metadataTypes": data.metadata_types,
        "minConsensusSize": data.min_consensus_size,
        "numberOfAssets": data.number_of_assets,
        "rules": data.rules,
        "numberOfSkippedAssets": data.number_of_skipped_assets,
        "numberOfRemainingAssets": data.number_of_remaining_assets,
        "numberOfReviewedAssets": data.number_of_reviewed_assets,
        "reviewCoverage": data.review_coverage,
        "shouldRelaunchKpiComputation": data.should_relaunch_kpi_computation,
        "title": data.title,
        "useHoneyPot": data.use_honeypot,
    }



"""Helpers for GraphQL Queries and Mutations."""

import functools
import glob
import mimetypes
import os
import re
import warnings
from json import dumps, loads
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union

import pyparsing as pp
import tenacity
from typing_extensions import get_args, get_origin

from kili.adapters.http_client import HttpClient
from kili.core.constants import mime_extensions_for_IV2

T = TypeVar("T")


def format_result(
    name: str, result: dict, object_: Optional[Type[T]], http_client: HttpClient
) -> T:
    """Formats the result of the GraphQL queries.

    Args:
        name: name of the field to extract, usually data
        result: query result to parse
        object_: returned type
        http_client: http client to use for the query
    """
    formatted_json = format_json(result[name], http_client)
    if object_ is None:
        return formatted_json  # pyright: ignore[reportGeneralTypeIssues]
    if isinstance(formatted_json, list):
        if get_origin(object_) is list:
            obj = get_args(object_)[0]
            # pylint: disable=line-too-long
            return [obj(element) for element in formatted_json]  # pyright: ignore[reportGeneralTypeIssues]
        # the legacy "orm" objects fall into this category.
        # pylint: disable=line-too-long
        return [object_(element) for element in formatted_json]  # pyright: ignore[reportGeneralTypeIssues]

    return object_(formatted_json)


def get_mime_type(path: str):
    """Provide the mime type of a file (e.g. image/png).

    Args:
        path: path of the file
    """
    path = path.lower()
    mime_type, _ = mimetypes.guess_type(path)

    # guess_type does not recognize JP2 files on Windows
    if mime_type is None and path.endswith(".jp2"):
        return "image/jp2"
    # guess_type provides a wrong mime type for NITF files on Ubuntu
    if path.endswith((".ntf", ".nitf")) and mime_type != "application/vnd.nitf":
        return "application/vnd.nitf"
    return mime_type


def is_url(path: object):
    """Check if the path is a url or something else.

    Args:
        path: path of the file
    """
    return isinstance(path, str) and re.match(r"^(http://|https://)", path.lower())


def __format_json_dict(result: Dict, http_client: HttpClient) -> Dict:
    """Parse a dictionary inside the result of a graphQL query to format json fields.

    If json fields (i.e "jsonInterface", "jsonMetadata" and "jsonResponse")
    are part of the keys, either:
        - fetch the json from the url if it is hosted on a bucket
        - load the json from the string if the json is given as a string

    Args:
        result: a dictionary included in the result of a GraphQL query
        http_client: http client to use for the query
    """
    for key, value in result.items():
        if key in ["jsonInterface", "jsonMetadata", "jsonResponse"]:  # TODO: also parse jsonContent
            if (value == "" or value is None) and not (is_url(value) and key == "jsonInterface"):
                result[key] = {}
            elif isinstance(value, str):
                try:
                    if is_url(value):
                        result[key] = http_client.get(value, timeout=30).json()
                    else:
                        result[key] = loads(value)
                except Exception as exception:
                    raise ValueError(
                        "Json Metadata / json response / json interface should be valid jsons"
                    ) from exception
        else:
            result[key] = format_json(value, http_client)
    return result


D = TypeVar("D")


def format_json(
    result: Union[None, list, dict, D], http_client: HttpClient
) -> Union[None, list, dict, D]:
    """Recusively parse the result of a GraphQL query in order to get json fields as a dictionary.

    Args:
        result: result of a GraphQL query
        http_client: http client to use for the query
    """
    if result is None:
        return result
    if isinstance(result, list):
        return [format_json(elem, http_client) for elem in result]
    if isinstance(result, dict):
        return __format_json_dict(result, http_client)
    return result


def deprecate(
    msg: Optional[str] = None,
    removed_in: Optional[str] = None,
    type_: Type[Warning] = DeprecationWarning,
):
    """Decorator factory that tag a deprecated function.

    - To deprecated the whole function, you can give a message at the decorator level.
    - For more sharp condition on the warning message, integrate this warning inside the function
    but still tag the function with this decorator, without giving a message argument

    Args:
        msg: string message that will be displayed whenever the function is called
        removed_in: string version in the format "Major.Minor"
            in which the deprecation element has to be removed
        type_: DeprecationWarning by default
    """

    def decorator(func):
        if removed_in:
            if len(removed_in.split(".")) != 2:
                raise ValueError(
                    f'"removed_in" argument in deprecate wrapper of the function {func.__name__}'
                    'should have the format "Major.Minor"'
                )
            func.removed_in = removed_in

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if msg:
                warnings.warn(msg, type_, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def format_metadata(metadata: object):
    """Formats metadata.

    Args:
        metadata: a python object
    """
    if metadata is None:
        return metadata
    if isinstance(metadata, str):
        return metadata
    if isinstance(metadata, (dict, list)):
        return dumps(metadata)
    raise TypeError(
        f"Metadata {metadata} of type {type(metadata)} must either be None,"
        " a string a list or a dict."
    )


def convert_to_list_of_none(array: List, length: int) -> List:
    """Turns a value in a list of length length.

    Args:
        array: the array to convert
        length: the length of the array
    """
    if isinstance(array, list):
        if len(array) != length:
            raise ValueError(f"array should have length {length}")
        return array
    return [None] * length


def is_none_or_empty(object_: object) -> bool:
    """Tests if an object is none or empty.

    Args:
        object_: a python object
    """
    object_is_empty = isinstance(object_, list) and len(object_) == 0
    return object_ is None or object_is_empty


def validate_category_search_query(query: str):
    """Validate the category search query.

    Args:
        query: the query to parse

    Raises:
        ValueError: if `query` is invalid
    """
    operator = pp.oneOf(">= <= > < ==")
    number = pp.pyparsing_common.number()
    dot = "."
    word = pp.Word(pp.alphas, pp.alphanums + "_-*")
    identifier = word + dot + word + dot + "count"
    condition = identifier + operator + number

    expr = pp.infixNotation(
        condition,
        [
            (
                "AND",
                2,
                pp.opAssoc.LEFT,
            ),
            (
                "OR",
                2,
                pp.opAssoc.LEFT,
            ),
        ],
    )
    try:
        expr.parseString(query, parseAll=True)
    except pp.ParseException as error:
        raise ValueError(f"Invalid category search query: {query}") from error


def get_file_paths_to_upload(
    files: List[str], file_check_function: Optional[Callable] = None, verbose: bool = False
) -> List[str]:
    """Get a list of paths for the files to upload given a list of files or folder paths.

    Args:
        files: a list path that can either be file paths, folder paths or unexisting paths
        file_check_function: function to check files. Must use argument path and return a bool
        verbose: if True, print the files that will be uploaded

    Returns:
        a list of the paths of the files to upload, compatible with the project type.
    """
    file_check_function = file_check_function or (lambda x: True)
    file_paths = []
    for item in files:
        if os.path.isfile(item):
            file_paths.append(item)
        elif os.path.isdir(item):
            folder_path = os.path.join(item, "")
            file_paths.extend(
                [sub_item for sub_item in glob.glob(folder_path + "*") if os.path.isfile(sub_item)]
            )
        else:
            file_paths.extend(
                [sub_item for sub_item in glob.glob(item) if os.path.isfile(sub_item)]
            )
    file_paths_to_upload = [path for path in file_paths if file_check_function(path)]
    if len(file_paths_to_upload) == 0:
        raise ValueError(
            "No files to upload. Check that the paths exist and that the file type is correct"
        )
    if verbose:
        for path in file_paths:
            if path not in file_paths_to_upload:
                print(f"{path:30} SKIPPED")
        if len(file_paths_to_upload) != len(file_paths):
            print("Paths skipped either do not exist or point towards an incorrect file")
    file_paths_to_upload.sort()
    return file_paths_to_upload


def check_file_mime_type(path: str, input_type: str, raise_error=True) -> bool:
    # pylint: disable=line-too-long
    """Returns true if the mime type of the file corresponds to the allowed mime types of the project."""
    mime_type = get_mime_type(path.lower())

    if not (mime_extensions_for_IV2[input_type] and mime_type):
        return False

    correct_mime_type = mime_type in mime_extensions_for_IV2[input_type]
    if not correct_mime_type and raise_error:
        raise ValueError(
            f"File mime type for {path} is {mime_type} and does not correspond "
            "to the type of the project. "
            f"File mime type should be one of {mime_extensions_for_IV2[input_type]}"
        )
    return correct_mime_type


class RetryLongWaitWarner:  # pylint: disable=too-few-public-methods
    """Class that warns when retry takes too long."""

    def __init__(
        self,
        warn_message: str,
        logger_func: Callable,
        warn_after: float = 10,
    ) -> None:
        """Class that warns when retry takes too long.

        Args:
            warn_message: custom warning message. If not provided, a default message is used.
            logger_func: function to log the message (print, warning, logger.warning, etc.)
            warn_after: time in seconds after which the warning is raised.
        """
        self.warn_message = warn_message
        self.logger_func = logger_func
        self.warn_after = warn_after

        self.warned = False

    def __call__(self, retry_state: tenacity.RetryCallState):
        if not self.warned and float(retry_state.outcome_timestamp or 0) > self.warn_after:
            self.logger_func(self.warn_message)
            self.warned = True


def is_empty_list_with_warning(method_name: str, argument_name: str, argument_value: Any) -> bool:
    """Check if an input list argument is empty and warn the user if it is.

    Returns True if the list is empty, False otherwise
    """
    if isinstance(argument_value, List) and len(argument_value) == 0:
        warnings.warn(
            f"Method '{method_name}' did nothing because the following argument"
            f" is empty: {argument_name}.",
            stacklevel=5,
        )
        return True
    return False



"""Common code for the yolo exporter."""

import json
import logging
from pathlib import Path
from typing import Dict, List, Optional, Union

from kili.services.export.exceptions import NotCompatibleInputType
from kili.services.export.format.base import AbstractExporter
from kili.services.types import Job


class LLMExporter(AbstractExporter):
    """Common code for LLM exporters."""

    def _check_arguments_compatibility(self) -> None:
        """Checks if the export label format is compatible with the export options."""
        if not self.with_assets:
            raise ValueError("LLM export format requires assets to be downloaded.")

    def _check_project_compatibility(self) -> None:
        """Checks if the export label format is compatible with the project."""
        if self.project["inputType"] != "LLM_RLHF":
            raise NotCompatibleInputType(
                f"Project with input type \"{self.project['inputType']}\" not compatible with LLM"
                " export format."
            )

    def _is_job_compatible(self, job: Job) -> bool:
        """Check job compatibility with the LLM format."""
        _ = job
        return True

    def _save_assets_export(self, assets: List[Dict], output_filename: Path) -> None:
        """Save the assets to a file and return the link to that file."""
        self.logger.info("Exporting to llm format...")

        export_json = json.dumps(assets, sort_keys=False, indent=4)
        with output_filename.open("wb") as output_file:
            output_file.write(export_json.encode("utf-8"))

    def process_and_save(
        self, assets: List[Dict], output_filename: Path
    ) -> Optional[List[Dict[str, Union[List[str], str]]]]:
        """LLM specific process and save."""
        result = self._process(assets)
        self._save_assets_export(result, output_filename)

    def process(self, assets: List[Dict]) -> List[Dict[str, Union[List[str], str]]]:
        """LLM specific process."""
        return self._process(assets)

    def _process(self, assets: List[Dict]) -> List[Dict[str, Union[List[str], str]]]:
        result = []
        for asset in assets:
            jobs_config = self.project["jsonInterface"]["jobs"]
            latest_label = asset["latestLabel"]
            result.append(
                {
                    "raw_data": _format_raw_data(asset),
                    "status": asset["status"],
                    "external_id": asset["externalId"],
                    "metadata": asset["jsonMetadata"],
                    "labels": [
                        {
                            "author": latest_label["author"]["email"],
                            "created_at": latest_label["createdAt"],
                            "label_type": latest_label["labelType"],
                            "label": _format_json_response(
                                jobs_config, latest_label["jsonResponse"]
                            ),
                        }
                    ],
                }
            )
        return result


def _format_json_response(
    jobs_config: Dict, json_response: Dict
) -> Dict[str, Union[str, List[str]]]:
    result = {}
    for job_name, job_value in json_response.items():
        job_config = jobs_config[job_name]
        if job_config is None:
            continue
        if job_config["mlTask"] == "CLASSIFICATION":
            result[job_name] = []
            for category in job_value["categories"]:
                result[job_name].append(category["name"])
                if "children" in category:
                    for child_name, child_value in _format_json_response(
                        jobs_config, category["children"]
                    ).items():
                        result[f"{job_name}.{category['name']}.{child_name}"] = child_value
        elif job_config["mlTask"] == "TRANSCRIPTION":
            result[job_name] = job_value["text"]
        else:
            logging.warning(f"Job {job_name} with mlTask {job_config['mlTask']} not supported")
    return result


def _format_raw_data(asset) -> List[Dict]:
    raw_data = []

    chat_id = asset["jsonMetadata"].get("chat_id", None)

    if (
        "chat_item_ids" in asset["jsonMetadata"]
        and isinstance(asset["jsonMetadata"]["chat_item_ids"], str)
        and len(asset["jsonMetadata"]["chat_item_ids"]) > 0
    ):
        chat_items_ids = str.split(asset["jsonMetadata"]["chat_item_ids"], "_")
    else:
        chat_items_ids = []

    if (
        "models" in asset["jsonMetadata"]
        and isinstance(asset["jsonMetadata"]["models"], str)
        and len(asset["jsonMetadata"]["models"]) > 0
    ):
        models = str.split(asset["jsonMetadata"]["models"], "_")
    else:
        models = []

    with Path(asset["content"]).open(encoding="utf8") as file:
        data = json.load(file)
    version = data.get("version", None)
    if version == "0.1":
        for index, prompt in enumerate(data["prompts"]):
            raw_data.append(
                {
                    "role": prompt.get("title", "user"),
                    "content": prompt["prompt"],
                    "id": _safe_pop(chat_items_ids),
                    "chat_id": chat_id,
                    "model": None,
                }
            )
            raw_data.extend(
                {
                    "role": completion.get("title", "assistant"),
                    "content": completion["content"],
                    "id": _safe_pop(chat_items_ids),
                    "chat_id": chat_id,
                    "model": _safe_pop(models) if index == len(data["prompts"]) - 1 else None,
                }
                for completion in prompt["completions"]
            )
    else:
        raise ValueError(f"Version {version} not supported")
    return raw_data


def _safe_pop(lst, index=0):
    try:
        return lst.pop(index)
    except IndexError:
        return None



"""Client presentation methods for notifications."""

from typing import Dict, Generator, Iterable, List, Literal, Optional, overload

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.domain.notification import NotificationFilter, NotificationId
from kili.domain.types import ListOrTuple
from kili.domain.user import UserFilter, UserId
from kili.presentation.client.helpers.common_validators import (
    disable_tqdm_if_as_generator,
)
from kili.use_cases.notification import NotificationUseCases
from kili.utils.logcontext import for_all_methods, log_call

from .base import BaseClientMethods


@for_all_methods(log_call, exclude=["__init__"])
class NotificationClientMethods(BaseClientMethods):
    """Methods attached to the Kili client, to run actions on notifications."""

    @overload
    def notifications(
        self,
        fields: ListOrTuple[str] = (
            "createdAt",
            "hasBeenSeen",
            "id",
            "message",
            "status",
            "userID",
        ),
        first: Optional[int] = None,
        has_been_seen: Optional[bool] = None,
        notification_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def notifications(
        self,
        fields: ListOrTuple[str] = (
            "createdAt",
            "hasBeenSeen",
            "id",
            "message",
            "status",
            "userID",
        ),
        first: Optional[int] = None,
        has_been_seen: Optional[bool] = None,
        notification_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def notifications(
        self,
        fields: ListOrTuple[str] = (
            "createdAt",
            "hasBeenSeen",
            "id",
            "message",
            "status",
            "userID",
        ),
        first: Optional[int] = None,
        has_been_seen: Optional[bool] = None,
        notification_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of notifications respecting a set of criteria.

        Args:
            fields: All the fields to request among the possible fields for the notifications
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#notification) for all possible fields.
            first: Number of notifications to query
            has_been_seen: If the notifications returned should have been seen.
            notification_id: If given, will return the notification which has this id
            skip: Number of notifications to skip (they are ordered by their date of creation,
                first to last).
            user_id: If given, returns the notifications of a specific user
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the notifications is returned.

        Returns:
            An iterable of notifications.
        """
        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)
        options = QueryOptions(disable_tqdm, first, skip)
        filters = NotificationFilter(
            has_been_seen=has_been_seen,
            id=NotificationId(notification_id) if notification_id else None,
            user=UserFilter(id=UserId(user_id)) if user_id else None,
        )
        notifications_gen = NotificationUseCases(self.kili_api_gateway).list_notifications(
            options=options, fields=fields, filters=filters
        )
        if as_generator:
            return notifications_gen
        return list(notifications_gen)

    @typechecked
    def count_notifications(
        self,
        has_been_seen: Optional[bool] = None,
        user_id: Optional[str] = None,
        notification_id: Optional[str] = None,
    ) -> int:
        """Count the number of notifications.

        Args:
            has_been_seen: Filter on notifications that have been seen.
            user_id: Filter on the notifications of a specific user.
            notification_id: Filter on a specific notification.

        Returns:
            The number of notifications with the parameters provided
        """
        filters = NotificationFilter(
            has_been_seen=has_been_seen,
            id=NotificationId(notification_id) if notification_id else None,
            user=UserFilter(id=UserId(user_id)) if user_id else None,
        )
        return NotificationUseCases(self.kili_api_gateway).count_notifications(filters=filters)



"""Tag use cases."""

from collections import defaultdict
from typing import Dict, List, Optional

from kili.adapters.kili_api_gateway.tag.types import UpdateTagReturnData
from kili.domain.project import ProjectId
from kili.domain.tag import TagId
from kili.domain.types import ListOrTuple
from kili.use_cases.base import BaseUseCases
from kili.utils.tqdm import tqdm


class TagUseCases(BaseUseCases):
    """Tag use cases."""

    def get_tags_of_organization(self, fields: ListOrTuple[str]) -> List[Dict]:
        """Get tags of organization."""
        return self._kili_api_gateway.list_tags_by_org(fields=fields)

    def get_tags_of_project(self, project_id: ProjectId, fields: ListOrTuple[str]) -> List[Dict]:
        """Get tags of project."""
        return self._kili_api_gateway.list_tags_by_project(project_id=project_id, fields=fields)

    def tag_project(
        self, project_id: ProjectId, tag_ids: ListOrTuple[TagId], disable_tqdm: Optional[bool]
    ) -> List[TagId]:
        """Assign tags to a project."""
        tags_of_orga = self._kili_api_gateway.list_tags_by_org(fields=("id",))
        tags_of_orga_ids = [tag["id"] for tag in tags_of_orga]

        tags_of_project = self._kili_api_gateway.list_tags_by_project(project_id, fields=("id",))
        tags_of_project_ids = [tag["id"] for tag in tags_of_project]

        ret_tags = []
        for tag_id in tqdm(tag_ids, desc="Tagging project", disable=disable_tqdm):
            if tag_id not in tags_of_orga_ids:
                raise ValueError(
                    f"Tag {tag_id} not found in organization with tag ids: {tags_of_orga}"
                )

            if tag_id in tags_of_project_ids:
                ret_tags.append(tag_id)
            else:
                ret_tags.append(
                    self._kili_api_gateway.check_tag(project_id=project_id, tag_id=tag_id)
                )

        return ret_tags

    def untag_project(
        self,
        project_id: ProjectId,
        tag_ids: ListOrTuple[TagId],
        disable_tqdm: Optional[bool],
    ) -> List[TagId]:
        """Remove tags from a project."""
        tag_ids_of_project = {
            tag["id"]
            for tag in self._kili_api_gateway.list_tags_by_project(
                project_id=ProjectId(project_id), fields=("id",)
            )
        }

        for tag in tag_ids:
            if tag not in tag_ids_of_project:
                raise ValueError(
                    f"Tag {tag} not found in project with tag ids: {tag_ids_of_project}"
                )

        return [
            self._kili_api_gateway.uncheck_tag(
                project_id=ProjectId(project_id), tag_id=TagId(tag_id)
            )
            for tag_id in tqdm(tag_ids, desc="Untagging project", disable=disable_tqdm)
        ]

    def get_tag_ids_from_labels(self, labels: ListOrTuple[str]) -> List[TagId]:
        """Get tag ids from labels."""
        tags_of_orga = self._kili_api_gateway.list_tags_by_org(fields=("id", "label"))

        tag_label_to_id = defaultdict(list)
        for tag in tags_of_orga:
            tag_label_to_id[tag["label"]].append(tag["id"])

        tag_ids = []
        for label in labels:
            if label not in tag_label_to_id:
                raise ValueError(
                    f"Tag `{label}` not found in organization with tag ids: {tags_of_orga}"
                )
            if len(tag_label_to_id[label]) > 1:
                raise ValueError(
                    f"Several tags with ids ({tag_label_to_id[label]}) have the same label"
                    f" `{label}`. Please use tag ids instead."
                )
            tag_ids.append(TagId(tag_label_to_id[label][0]))

        return tag_ids

    def update_tag(self, tag_id: TagId, new_tag_name: str) -> UpdateTagReturnData:
        """Update tag.

        This operation is organization-wide.
        The tag will be updated for all projects of the organization.

        Args:
            tag_id: Id of the tag to update.
            new_tag_name: New name of the tag.

        Returns:
            The updated tag.
        """
        return self._kili_api_gateway.update_tag(tag_id=tag_id, label=new_tag_name)

    def delete_tag(self, tag_id: TagId) -> bool:
        """Delete the given tag.

        This operation is organization-wide.
        The tag will no longer be proposed for projects of the organization.
        If this tag is checked for one or more projects of the organization, it will be unchecked.

        Args:
            tag_id: Id of the tag to remove.

        Returns:
            Whether the tag was successfully removed.
        """
        return self._kili_api_gateway.delete_tag(tag_id=tag_id)



"""Code specific to Azure blob storage."""

import re
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlparse

from azure.storage.blob import BlobServiceClient

from kili.domain.project import InputType


def get_blob_paths_azure_data_connection_with_service_credentials(
    data_connection: Dict, data_integration: Dict, input_type: InputType
) -> Tuple[List[str], List[str], List[str]]:
    """Get the blob paths for an Azure data connection using service credentials."""
    if not (data_integration["azureSASToken"] and data_integration["azureConnectionURL"]):
        raise ValueError(
            f"Cannot retrieve blob paths for data connection {data_connection['id']} with data"
            f" integration {data_integration['id']}. Need to provide \"azureSASToken\" and"
            f' "azureConnectionURL" in data integration: {data_integration}'
        )

    return AzureBucket(
        sas_token=data_integration["azureSASToken"],
        connection_url=data_integration["azureConnectionURL"],
    ).get_blob_paths_azure_data_connection_with_service_credentials(
        input_type=input_type,
        selected_folders=data_connection.get("selectedFolders"),
        prefix=data_connection.get("prefix"),
        include=data_connection.get("include"),
        exclude=data_connection.get("exclude"),
    )


class AzureBucket:
    """Class for Azure blob storage buckets."""

    def __init__(self, sas_token: str, connection_url: str) -> None:
        """Initialize the Azure bucket."""
        self.sas_token = sas_token
        self.connection_url = connection_url

        (
            self.storage_account,
            self.container_name,
        ) = self._split_connection_url_into_storage_account_and_container_name(connection_url)
        self.blob_sas_url = f"https://{self.storage_account}.blob.core.windows.net?{self.sas_token}"
        self.client = BlobServiceClient(account_url=self.blob_sas_url)
        self.storage_bucket = self.client.get_container_client(self.container_name)

    @staticmethod
    def _split_connection_url_into_storage_account_and_container_name(
        connection_url: str,
    ) -> Tuple[str, str]:
        """Split the connection url into storage account and container name."""
        split_value = ".blob.core.windows.net"
        url_connection = urlparse(connection_url)
        # pylint: disable=line-too-long
        storage_account = url_connection.hostname.split(  # pyright: ignore[reportOptionalMemberAccess]
            split_value
        )[0]
        container_name = url_connection.path.lstrip("/")
        return storage_account, container_name

    def get_blob_paths_as_tree(self) -> Dict:
        """Get a tree representation of the Azure bucket.

        Folder structure is represented as a dictionary.
        A folder is represented as a key with a value as a dictionary.
        A file is represented as a key with a value as None.
        """
        filetree = {}

        for filepath in self.storage_bucket.list_blob_names():
            current_node = filetree
            *folders, filename = Path(filepath).parts
            for folder in folders:
                if folder not in current_node:
                    current_node[folder] = {}
                current_node = current_node[folder]
            if filename:
                current_node[filename] = None

        return filetree

    @staticmethod
    def _generate_regex_pattern(pattern: str) -> str:
        pattern = re.escape(pattern)

        if "*" in pattern:
            pattern = pattern.replace(r"\*", ".+")
            pattern = f"^{pattern}$"
        else:
            pattern = f"^{pattern}"

        return pattern

    def get_blob_paths_azure_data_connection_with_service_credentials(
        self,
        input_type: InputType,
        selected_folders: Optional[List[str]] = None,
        prefix: Optional[str] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ) -> Tuple[List[str], List[str], List[str]]:
        """Get the blob paths for an Azure data connection using service credentials."""
        blob_paths = []
        content_types = []
        warnings = set()
        include = [self._generate_regex_pattern(pattern) for pattern in include] if include else []
        exclude = [self._generate_regex_pattern(pattern) for pattern in exclude] if exclude else []
        for blob in self.storage_bucket.list_blobs(name_starts_with=prefix):
            if not hasattr(blob, "name") or not isinstance(blob.name, str):
                continue

            if (
                prefix is None
                and isinstance(selected_folders, List)
                and not any(
                    blob.name.startswith(selected_folder) for selected_folder in selected_folders
                )
            ):
                continue

            if len(include) > 0 and not any(re.match(pattern, blob.name) for pattern in include):
                continue

            if len(exclude) > 0 and any(re.match(pattern, blob.name) for pattern in exclude):
                continue

            has_content_type_field = (
                hasattr(blob, "content_settings")
                and hasattr(blob.content_settings, "content_type")
                and isinstance(blob.content_settings.content_type, str)
            )
            if not has_content_type_field:
                warnings.add("Objects with missing content-type were ignored")

            elif not self._is_content_type_compatible_with_input_type(
                blob.content_settings.content_type,  # pyright: ignore[reportGeneralTypeIssues]
                input_type,
            ):
                warnings.add(
                    "Objects with unsupported content-type for this type of project were ignored"
                )

            else:
                blob_paths.append(blob.name)
                content_types.append(blob.content_settings.content_type)

        return blob_paths, list(warnings), content_types

    @staticmethod
    def _is_content_type_compatible_with_input_type(
        content_type: str, input_type: InputType
    ) -> bool:
        """Check if the content type is compatible with the input type."""
        if input_type == "IMAGE":
            return content_type.startswith("image")

        if input_type == "VIDEO":
            return content_type.startswith("video")

        if input_type == "PDF":
            return content_type.startswith("application/pdf")

        if input_type == "TEXT":
            return any(
                content_type.startswith(text_type)
                for text_type in (
                    "application/json",
                    "text/plain",
                    "text/html",
                    "text/csv",
                    "text/xml",
                )
            )

        raise ValueError(f"Unknown project input type: {input_type}")



"""Mixin extending Kili API Gateway class with label related operations."""

import json
from typing import Dict, Generator, List, Optional

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.adapters.kili_api_gateway.project.common import get_project
from kili.core.constants import MUTATION_BATCH_SIZE
from kili.core.utils.pagination import batcher
from kili.domain.asset import AssetId
from kili.domain.label import LabelFilters, LabelId
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.utils.tqdm import tqdm

from .annotation_to_json_response import (
    AnnotationsToJsonResponseConverter,
)
from .common import get_annotation_fragment
from .formatters import load_label_json_fields
from .mappers import (
    append_label_data_mapper,
    append_to_labels_data_mapper,
    label_where_mapper,
    update_label_data_mapper,
)
from .operations import (
    GQL_COUNT_LABELS,
    GQL_DELETE_LABELS,
    get_append_many_labels_mutation,
    get_append_to_labels_mutation,
    get_create_honeypot_mutation,
    get_labels_query,
    get_update_properties_in_label_mutation,
)
from .types import AppendManyLabelsData, AppendToLabelsData, UpdateLabelData


class LabelOperationMixin(BaseOperationMixin):
    """Mixin extending Kili API Gateway class with label related operations."""

    def count_labels(self, filters: LabelFilters) -> int:
        """Count labels."""
        variables = {"where": label_where_mapper(filters)}
        result = self.graphql_client.execute(GQL_COUNT_LABELS, variables)
        return result["data"]

    def list_labels(
        self,
        filters: LabelFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """List labels."""
        if "jsonResponse" in fields:
            project_info = get_project(
                self.graphql_client, filters.project_id, ("inputType", "jsonInterface")
            )
            if project_info["inputType"] in {"VIDEO", "LLM_RLHF"}:
                yield from self.list_labels_split(filters, fields, options, project_info)
                return

        fragment = fragment_builder(fields)
        query = get_labels_query(fragment)
        where = label_where_mapper(filters)
        labels_gen = PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving labels", GQL_COUNT_LABELS
        )
        labels_gen = (load_label_json_fields(label, fields) for label in labels_gen)
        yield from labels_gen

    def list_labels_split(
        self, filters: LabelFilters, fields: ListOrTuple[str], options: QueryOptions, project_info
    ) -> Generator[Dict, None, None]:
        """List labels."""
        if project_info["inputType"] == "VIDEO":
            options = QueryOptions(
                options.disable_tqdm, options.first, options.skip, min(options.batch_size, 20)
            )

        fragment = fragment_builder(fields)
        inner_annotation_fragment = get_annotation_fragment()
        full_fragment = f"""
            {fragment}
            annotations {{
                {inner_annotation_fragment}
            }}
        """
        query = get_labels_query(full_fragment)
        where = label_where_mapper(filters)
        labels_gen = PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving labels", GQL_COUNT_LABELS
        )
        labels_gen = (load_label_json_fields(label, fields) for label in labels_gen)

        if "jsonResponse" in fields:
            converter = AnnotationsToJsonResponseConverter(
                json_interface=project_info["jsonInterface"],
                project_input_type=project_info["inputType"],
            )
            for label in labels_gen:
                converter.patch_label_json_response(label, label["annotations"])
                if "annotations" not in fields:
                    label.pop("annotations")
                yield label

        else:
            yield from labels_gen

    def update_properties_in_label(
        self, label_id: LabelId, data: UpdateLabelData, fields: ListOrTuple[str]
    ) -> Dict:
        """Update properties in label."""
        fragment = fragment_builder(fields)
        query = get_update_properties_in_label_mutation(fragment)
        variables = {"where": {"id": label_id}, "data": update_label_data_mapper(data)}
        result = self.graphql_client.execute(query, variables)
        modified_label = result["data"]
        return load_label_json_fields(modified_label, fields=fields)

    def delete_labels(
        self, ids: ListOrTuple[LabelId], disable_tqdm: Optional[bool]
    ) -> List[LabelId]:
        """Delete labels."""
        delete_label_ids: List[LabelId] = []

        with tqdm(total=len(ids), desc="Deleting labels", disable=disable_tqdm) as pbar:
            for batch_of_label_ids in batcher(ids, batch_size=MUTATION_BATCH_SIZE):
                variables = {"ids": batch_of_label_ids}
                result = self.graphql_client.execute(GQL_DELETE_LABELS, variables)
                batch_deleted_label_ids = result["data"]
                delete_label_ids.extend(batch_deleted_label_ids)
                pbar.update(len(batch_of_label_ids))

        return delete_label_ids

    def append_many_labels(
        self,
        data: AppendManyLabelsData,
        fields: ListOrTuple[str],
        disable_tqdm: Optional[bool],
        project_id: Optional[ProjectId],
    ) -> List[Dict]:
        """Append many labels."""
        nb_labels_to_add = len(data.labels_data)

        fragment = fragment_builder(fields)
        query = get_append_many_labels_mutation(fragment=fragment)

        added_labels: List[Dict] = []
        with tqdm(total=nb_labels_to_add, desc="Adding labels", disable=disable_tqdm) as pbar:
            for batch_of_label_data in batcher(data.labels_data, batch_size=MUTATION_BATCH_SIZE):
                variables = {
                    "data": {
                        "labelType": data.label_type,
                        "overwrite": data.overwrite,
                        "labelsData": [
                            append_label_data_mapper(label) for label in batch_of_label_data
                        ],
                    },
                    "where": {
                        "idIn": [label.asset_id for label in batch_of_label_data],
                    },
                }
                if project_id is not None:
                    variables["where"]["project"] = {"id": project_id}

                # we increase the timeout because the import can take a long time
                batch_result = self.graphql_client.execute(query, variables, timeout=120)
                added_labels.extend(batch_result["data"])
                pbar.update(len(batch_of_label_data))

        return added_labels

    def append_to_labels(
        self, data: AppendToLabelsData, asset_id: AssetId, fields: ListOrTuple[str]
    ) -> Dict:
        """Append to labels."""
        fragment = fragment_builder(fields)
        query = get_append_to_labels_mutation(fragment)
        variables = {
            "where": {"id": asset_id},
            "data": append_to_labels_data_mapper(data),
        }
        result = self.graphql_client.execute(query, variables)
        return result["data"]

    def create_honeypot_label(
        self, json_response: Dict, asset_id: AssetId, fields: ListOrTuple[str]
    ) -> Dict:
        """Create honeypot label."""
        fragment = fragment_builder(fields)
        query = get_create_honeypot_mutation(fragment)
        variables = {
            "data": {"jsonResponse": json.dumps(json_response)},
            "where": {"id": asset_id},
        }
        result = self.graphql_client.execute(query, variables)
        return result["data"]



"""Project gateway common."""
from typing import Dict

from kili.adapters.kili_api_gateway.helpers.queries import (
    fragment_builder,
)
from kili.adapters.kili_api_gateway.project.formatters import (
    load_project_json_fields,
)
from kili.adapters.kili_api_gateway.project.operations import get_projects_query
from kili.core.graphql.graphql_client import GraphQLClient
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.exceptions import NotFound


def get_project(
    graphql_client: GraphQLClient, project_id: ProjectId, fields: ListOrTuple[str]
) -> Dict:
    """Get project."""
    fragment = fragment_builder(fields)
    query = get_projects_query(fragment)
    result = graphql_client.execute(
        query=query, variables={"where": {"id": project_id}, "first": 1, "skip": 0}
    )
    projects = result["data"]

    if len(projects) == 0:
        raise NotFound(
            f"project ID: {project_id}. The project does not exist or you do not have access"
            " to it."
        )
    return load_project_json_fields(projects[0], fields)



"""Service for importing objects into kili."""

from typing import TYPE_CHECKING, Dict, List, Optional, Type, cast

from kili.domain.project import ProjectId
from kili.services.asset_import.exceptions import (
    ImportValidationError,
)

from .base import (
    BaseAbstractAssetImporter,
    LoggerParams,
    ProcessingParams,
    ProjectParams,
)
from .image import ImageDataImporter
from .llm import LLMDataImporter
from .pdf import PdfDataImporter
from .text import TextDataImporter
from .types import AssetLike
from .video import VideoDataImporter

if TYPE_CHECKING:
    from kili.client import Kili

importer_by_type: Dict[str, Type[BaseAbstractAssetImporter]] = {
    "PDF": PdfDataImporter,
    "IMAGE": ImageDataImporter,
    "TEXT": TextDataImporter,
    "VIDEO": VideoDataImporter,
    "VIDEO_LEGACY": VideoDataImporter,
    "LLM_RLHF": LLMDataImporter,
}


def import_assets(  # pylint: disable=too-many-arguments
    kili: "Kili",
    project_id: ProjectId,
    assets: List[Dict],
    raise_error: bool = True,
    disable_tqdm: Optional[bool] = False,
    verify: bool = True,
):
    """Import the selected assets into the specified project."""
    input_type = kili.kili_api_gateway.get_project(project_id, ("inputType",))["inputType"]

    project_params = ProjectParams(project_id=project_id, input_type=input_type)
    processing_params = ProcessingParams(raise_error=raise_error, verify=verify)
    logger_params = LoggerParams(disable_tqdm=disable_tqdm)
    importer_params = (kili, project_params, processing_params, logger_params)

    if input_type not in importer_by_type:
        raise NotImplementedError(f"There is no imported for the input type: {input_type}")
    if input_type != "IMAGE" and any(asset.get("multi_layer_content") for asset in assets):
        raise ImportValidationError(
            f"Import of multi-layer assets is not supported for input type: {input_type}"
        )
    asset_importer = importer_by_type[input_type](*importer_params)
    casted_assets = cast(List[AssetLike], assets)
    asset_importer.check_asset_contents(casted_assets)
    return asset_importer.import_assets(assets=casted_assets)



"""This script lists package constants."""


mime_extensions = {
    "Audio": "audio/x-flac,audio/mpeg,video/mp4",
    "Csv": "text/csv",
    "Video": (
        "video/mp4,video/x-matroska,video/3gpp,video/x-msvideo,"
        "video/x-m4v,video/quicktime,video/webm"
    ),
    "Image": (
        "application/vnd.nitf, image/jp2,image/jpeg,image/png,image/bmp,image/gif,image/webp,"
        "image/x-icon,image/tiff,image/vnd.microsoft.icon,image/svg+xml,image/avif,image/apng"
    ),
    "Pdf": "application/pdf",
    "Text": "text/plain",
    "TimeSeries": "text/csv",
    "LLM": "application/json",
}

mime_extensions_for_IV2 = {
    "AUDIO": mime_extensions["Audio"],
    "IMAGE": mime_extensions["Image"],
    "NA": "",
    "PDF": mime_extensions["Pdf"],
    "TEXT": mime_extensions["Text"],
    "TIME_SERIES": mime_extensions["Csv"],
    "URL": "",
    "VIDEO": mime_extensions["Video"],
    "VIDEO_LEGACY": "",
    "LLM_RLHF": mime_extensions["LLM"],
}

mime_extensions_for_py_scripts = ["text/x-python"]
mime_extensions_for_txt_files = ["text/plain"]

# These extensions require post-processing either because they can contain geospatial metadata,
# or because they are not directly supported by the browser.
mime_extensions_that_need_post_processing = ["application/vnd.nitf", "image/jp2", "image/tiff"]

QUERY_BATCH_SIZE = 100
MUTATION_BATCH_SIZE = 100
MAX_CALLS_PER_MINUTE = 500



"""Mappers for notification API calls."""

from typing import Dict

from kili.adapters.kili_api_gateway.user.mappers import user_where_mapper
from kili.domain.notification import NotificationFilter


def map_notification_filter(filters: NotificationFilter) -> Dict:
    """Build the GraphQL NotificationWhere variable to be sent in an operation."""
    return {
        "hasBeenSeen": filters.has_been_seen,
        "id": filters.id,
        "user": user_where_mapper(filters.user) if filters.user else None,
    }



"""GraphQL payload data mappers for label operations."""

import json
from typing import Dict

from kili.adapters.kili_api_gateway.asset.mappers import asset_where_mapper
from kili.adapters.kili_api_gateway.user.mappers import user_where_mapper
from kili.domain.label import LabelFilters

from .types import AppendLabelData, AppendToLabelsData, UpdateLabelData


def label_where_mapper(filters: LabelFilters) -> Dict[str, object]:
    """Map label filters to GraphQL LabelWhere."""
    return {
        "asset": asset_where_mapper(filters.asset) if filters.asset else None,
        "authorIn": filters.author_in,
        "consensusMarkGte": filters.consensus_mark_gte,
        "consensusMarkLte": filters.consensus_mark_lte,
        "createdAt": filters.created_at,
        "createdAtGte": filters.created_at_gte,
        "createdAtLte": filters.created_at_lte,
        "honeypotMarkGte": filters.honeypot_mark_gte,
        "honeypotMarkLte": filters.honeypot_mark_lte,
        "id": filters.id,
        "idIn": filters.id_in,
        "labelerIn": filters.labeler_in,
        "project": {"id": filters.project_id},
        "reviewerIn": filters.reviewer_in,
        "search": filters.search,
        "typeIn": filters.type_in,
        "user": user_where_mapper(filters.user) if filters.user else None,
    }


def update_label_data_mapper(data: UpdateLabelData) -> Dict:
    """Map UpdateLabelData to GraphQL LabelData input."""
    return {
        "isSentBackToQueue": data.is_sent_back_to_queue,
        "jsonResponse": json.dumps(data.json_response) if data.json_response else None,
        "modelName": data.model_name,
        "secondsToLabel": data.seconds_to_label,
    }


def append_label_data_mapper(data: AppendLabelData) -> Dict:
    """Map AppendLabelData to GraphQL AppendLabelData input."""
    return {
        "authorID": data.author_id,
        "assetID": data.asset_id,
        "clientVersion": data.client_version,
        "jsonResponse": json.dumps(data.json_response),
        "secondsToLabel": data.seconds_to_label,
        "modelName": data.model_name,
    }


def append_to_labels_data_mapper(data: AppendToLabelsData) -> Dict:
    """Map AppendToLabelsData to GraphQL AppendToLabelsData input."""
    return {
        "authorID": data.author_id,
        "clientVersion": data.client_version,
        "jsonResponse": json.dumps(data.json_response),
        "labelType": data.label_type,
        "secondsToLabel": data.seconds_to_label,
        "skipped": data.skipped,
    }



"""Annotation object to json response converter."""
import math
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Generator, List, Optional, Tuple, TypeVar, Union, cast, overload

from kili.domain.annotation import (
    ClassicAnnotation,
    ClassificationAnnotation,
    RankingAnnotation,
    TranscriptionAnnotation,
    Vertice,
    VideoAnnotation,
    VideoClassificationAnnotation,
    VideoClassificationKeyAnnotation,
    VideoObjectDetectionAnnotation,
    VideoObjectDetectionKeyAnnotation,
    VideoTranscriptionAnnotation,
    VideoTranscriptionKeyAnnotation,
)
from kili.domain.ontology import JobName, JobTool


class AnnotationsToJsonResponseConverter:
    """Convert annotations to JSON response."""

    def __init__(self, project_input_type: str, json_interface) -> None:
        """Initialize the converter."""
        self._project_input_type = project_input_type
        self._project_json_interface = json_interface

    def _label_has_json_response_data(self, label: Dict) -> bool:
        if self._project_input_type == "VIDEO":
            job_names_in_json_resp = {
                job_name for frame_resp in label["jsonResponse"].values() for job_name in frame_resp
            }
        else:
            job_names_in_json_resp = set(label["jsonResponse"].keys())

        if any(
            job_name in job_names_in_json_resp for job_name in self._project_json_interface["jobs"]
        ):
            return True

        return False

    def patch_label_json_response(
        self, label: Dict, annotations: Union[List[VideoAnnotation], List[ClassicAnnotation]]
    ) -> None:
        """Patch the label json response using the annotations.

        Modifies the input label.
        """
        if self._project_input_type in {"VIDEO", "LLM_RLHF"}:
            if not annotations and self._label_has_json_response_data(label):
                return

            if self._project_input_type == "VIDEO":
                annotations = cast(List[VideoAnnotation], annotations)
                converted_json_resp = _video_annotations_to_json_response(
                    annotations=annotations, json_interface=self._project_json_interface
                )
            else:
                annotations = cast(List[ClassicAnnotation], annotations)
                converted_json_resp = _classic_annotations_to_json_response(annotations=annotations)

            label["jsonResponse"] = converted_json_resp


def _add_annotation_metadata(annotations: List[VideoAnnotation], json_response: Dict) -> None:
    for ann in annotations:
        if ann["__typename"] == "VideoObjectDetectionAnnotation":
            ann = cast(VideoObjectDetectionAnnotation, ann)
            job_counter = (
                json_response["0"]
                .setdefault("ANNOTATION_JOB_COUNTER", {})
                .setdefault(ann["job"], defaultdict(int))
            )
            job_counter[ann["category"]] += 1
            json_response["0"].setdefault("ANNOTATION_NAMES_JOB", {})[ann["mid"]] = ann["name"]


def _fill_empty_frames(json_response: Dict) -> None:
    max_frame_id = max((int(frame_id) for frame_id in json_response), default=-1)
    for frame_id in range(max_frame_id + 1):
        json_response.setdefault(str(frame_id), {})


def _video_annotations_to_json_response(
    annotations: List[VideoAnnotation], json_interface: Dict
) -> Dict[str, Dict[JobName, Dict]]:
    """Convert video label annotations to a video json response."""
    json_resp = defaultdict(dict)

    for i, ann in enumerate(annotations):
        if ann["path"]:  # skip child annotations
            continue

        other_annotations = annotations[:i] + annotations[i + 1 :]

        if ann["__typename"] == "VideoObjectDetectionAnnotation":
            ann = cast(VideoObjectDetectionAnnotation, ann)
            ann_json_resp = _video_object_detection_annotation_to_json_response(
                ann, other_annotations, json_interface=json_interface
            )
            for frame_id, frame_json_resp in ann_json_resp.items():
                for job_name, job_resp in frame_json_resp.items():
                    json_resp[frame_id].setdefault(job_name, {}).setdefault(
                        "annotations", []
                    ).extend(job_resp["annotations"])

        elif ann["__typename"] == "VideoClassificationAnnotation":
            ann = cast(VideoClassificationAnnotation, ann)
            ann_json_resp = _video_classification_annotation_to_json_response(
                ann, other_annotations
            )
            for frame_id, frame_json_resp in ann_json_resp.items():
                for job_name, job_resp in frame_json_resp.items():
                    json_resp[frame_id].setdefault(job_name, {}).setdefault(
                        "categories", []
                    ).extend(job_resp["categories"])
                    json_resp[frame_id][job_name]["isKeyFrame"] = job_resp["isKeyFrame"]

        elif ann["__typename"] == "VideoTranscriptionAnnotation":
            ann = cast(VideoTranscriptionAnnotation, ann)
            ann_json_resp = _video_transcription_annotation_to_json_response(ann)
            for frame_id, frame_json_resp in ann_json_resp.items():
                json_resp[frame_id] = {**json_resp[frame_id], **frame_json_resp}

        else:
            raise NotImplementedError(f"Cannot convert video annotation to json response: {ann}")

    _add_annotation_metadata(annotations, json_resp)
    _fill_empty_frames(json_resp)

    return dict(sorted(json_resp.items(), key=lambda item: int(item[0])))  # sort by frame id


def _classic_annotations_to_json_response(
    annotations: List[ClassicAnnotation],
) -> Dict[str, Dict[JobName, Dict]]:
    """Convert label annotations to a json response."""
    json_resp = defaultdict(dict)

    for i, ann in enumerate(annotations):
        if ann["path"]:  # skip child annotations
            continue

        other_annotations = annotations[:i] + annotations[i + 1 :]

        if ann["__typename"] == "ClassificationAnnotation":
            ann = cast(ClassificationAnnotation, ann)
            ann_json_resp = _classification_annotation_to_json_response(ann, other_annotations)
            for job_name, job_resp in ann_json_resp.items():
                json_resp.setdefault(job_name, {}).setdefault("categories", []).extend(
                    job_resp["categories"]
                )

        elif ann["__typename"] == "RankingAnnotation":
            ann = cast(RankingAnnotation, ann)
            ann_json_resp = _ranking_annotation_to_json_response(ann)
            for job_name, job_resp in ann_json_resp.items():
                json_resp.setdefault(job_name, {}).setdefault("orders", []).extend(
                    job_resp["orders"]
                )

        elif ann["__typename"] == "TranscriptionAnnotation":
            ann = cast(TranscriptionAnnotation, ann)
            ann_json_resp = _transcription_annotation_to_json_response(ann)
            for job_name, job_resp in ann_json_resp.items():
                json_resp.setdefault(job_name, {}).setdefault("text", job_resp["text"])

        else:
            raise NotImplementedError(f"Cannot convert classic annotation to json response: {ann}")

    return dict(json_resp)


@overload
def _key_annotations_iterator(
    annotation: VideoTranscriptionAnnotation,
) -> Generator[
    Tuple[
        VideoTranscriptionKeyAnnotation,
        int,
        int,
        Optional[VideoTranscriptionKeyAnnotation],
        Optional[int],
    ],
    None,
    None,
]:
    ...


@overload
def _key_annotations_iterator(
    annotation: VideoClassificationAnnotation,
) -> Generator[
    Tuple[
        VideoClassificationKeyAnnotation,
        int,
        int,
        Optional[VideoClassificationKeyAnnotation],
        Optional[int],
    ],
    None,
    None,
]:
    ...


@overload
def _key_annotations_iterator(
    annotation: VideoObjectDetectionAnnotation,
) -> Generator[
    Tuple[
        VideoObjectDetectionKeyAnnotation,
        int,
        int,
        Optional[VideoObjectDetectionKeyAnnotation],
        Optional[int],
    ],
    None,
    None,
]:
    ...


def _key_annotations_iterator(annotation: VideoAnnotation) -> Generator:
    """Helper to iterate over the key annotations of a video annotation.

    The key annotations are sorted by frame id.
    """
    sorted_key_annotations = sorted(
        annotation["keyAnnotations"], key=lambda key_ann: int(key_ann["frame"])
    )
    # previous_key_ann is used to keep track of the previous key annotation
    # in case where keyframe is not present in the current frame range
    previous_key_ann = {}
    previous_key_ann_index = 0
    # iterate over the frame ranges of the annotation
    for frame_interval in annotation["frames"]:
        frame_range = range(frame_interval["start"], frame_interval["end"] + 1)
        # has_key_annotation is used to keep track of whether the current frame range
        # has a key annotation or not. It could be that the current frame range does not
        # have a key annotation, or that the first keyframe is after the start of the frame range
        has_key_annotation = False
        for key_ann_index, key_ann in enumerate(sorted_key_annotations):
            # skip the key annotation if the key annotation start frame
            # is not in current frame range
            if key_ann["frame"] not in frame_range:
                continue

            if key_ann["frame"] > frame_interval["start"] and not has_key_annotation:
                # if the key annotation start frame is after the start of the frame range,
                # then we need to yield the previous key annotation
                key_ann_frame = previous_key_ann["frame"]
                yield (
                    previous_key_ann,
                    frame_interval["start"],
                    key_ann["frame"],
                    key_ann,
                    key_ann_frame,
                )
            # compute the key annotation frame range
            # the start frame of key annotation is given, but not the end frame
            key_ann_start = key_ann["frame"]
            key_ann_frame = key_ann["frame"]
            key_ann_end = min(
                frame_interval["end"] + 1,
                sorted_key_annotations[key_ann_index + 1]["frame"]
                if key_ann_index + 1 < len(sorted_key_annotations)
                else frame_interval["end"] + 1,
            )

            # get the next key annotation, if it exists
            next_key_ann = (
                sorted_key_annotations[key_ann_index + 1]
                if key_ann_index + 1 < len(sorted_key_annotations)
                else None
            )

            has_key_annotation = True
            previous_key_ann = key_ann
            previous_key_ann_index = key_ann_index

            yield key_ann, key_ann_start, key_ann_end, next_key_ann, key_ann_frame

        if not has_key_annotation:
            key_ann_frame = previous_key_ann["frame"]
            next_key_ann = (
                sorted_key_annotations[previous_key_ann_index + 1]
                if previous_key_ann_index + 1 < len(sorted_key_annotations)
                else None
            )
            yield (
                previous_key_ann,
                frame_interval["start"],
                frame_interval["end"] + 1,
                next_key_ann,
                key_ann_frame,
            )


def _ranking_annotation_to_json_response(
    annotation: RankingAnnotation,
) -> Dict[JobName, Dict]:
    """Convert ranking annotation to a json response.

    Ranking jobs cannot have child jobs.
    """
    json_resp = {
        annotation["job"]: {
            "orders": sorted(
                annotation["annotationValue"]["orders"], key=lambda item: int(item["rank"])
            ),
        }
    }

    return json_resp


def _transcription_annotation_to_json_response(
    annotation: TranscriptionAnnotation,
) -> Dict[JobName, Dict]:
    """Convert transcription annotation to a json response.

    Transcription jobs cannot have child jobs.
    """
    json_resp = {
        annotation["job"]: {
            "text": annotation["annotationValue"]["text"],
        }
    }

    return json_resp


def _video_transcription_annotation_to_json_response(
    annotation: VideoTranscriptionAnnotation,
) -> Dict[str, Dict[JobName, Dict]]:
    """Convert video transcription annotation to a json response.

    Transcription jobs cannot have child jobs.
    """
    json_resp: Dict[str, Dict[JobName, Dict]] = defaultdict(dict)

    for key_ann, key_ann_start, key_ann_end, _, key_ann_frame in _key_annotations_iterator(
        annotation
    ):
        for frame_id in range(key_ann_start, key_ann_end):
            json_resp[str(frame_id)][annotation["job"]] = {
                "isKeyFrame": frame_id == key_ann_frame,
                "text": key_ann["annotationValue"]["text"],
            }

    return json_resp


T = TypeVar("T", VideoAnnotation, ClassicAnnotation)


def _get_child_annotations(annotation: T, other_annotations: List[T]) -> List[T]:
    """Get the child annotations (child jobs) of a video annotation."""
    return [
        ann
        for ann in other_annotations
        # ann["path"] is a list of couples (annotationId, category)
        if len(ann["path"]) > 0
        and ann["path"][-1][0] == annotation["id"]
        and annotation["path"] == ann["path"][:-1]
    ]


def _compute_children_json_resp(
    child_annotations: Union[List[ClassicAnnotation], List[VideoAnnotation]],
    other_annotations: Union[List[ClassicAnnotation], List[VideoAnnotation]],
) -> Dict[str, Dict[JobName, Dict]]:
    """Compute the video json response of the child jobs of a video annotation."""
    children_json_resp = defaultdict(dict)

    for child_ann in child_annotations:
        if child_ann["__typename"] == "ClassificationAnnotation":
            child_ann = cast(ClassificationAnnotation, child_ann)
            other_annotations = cast(List[ClassicAnnotation], other_annotations)
            sub_job_resp = _classification_annotation_to_json_response(
                child_ann, _get_child_annotations(child_ann, other_annotations)
            )

        elif child_ann["__typename"] == "RankingAnnotation":
            child_ann = cast(RankingAnnotation, child_ann)
            sub_job_resp = _ranking_annotation_to_json_response(child_ann)

        elif child_ann["__typename"] == "TranscriptionAnnotation":
            child_ann = cast(TranscriptionAnnotation, child_ann)
            sub_job_resp = _transcription_annotation_to_json_response(child_ann)

        elif child_ann["__typename"] == "VideoClassificationAnnotation":
            child_ann = cast(VideoClassificationAnnotation, child_ann)
            other_annotations = cast(List[VideoAnnotation], other_annotations)
            sub_job_resp = _video_classification_annotation_to_json_response(
                child_ann, _get_child_annotations(child_ann, other_annotations)
            )

        elif child_ann["__typename"] == "VideoTranscriptionAnnotation":
            child_ann = cast(VideoTranscriptionAnnotation, child_ann)
            sub_job_resp = _video_transcription_annotation_to_json_response(child_ann)

        else:
            raise NotImplementedError(
                f"Cannot convert child annotation to json response: {child_ann}"
            )

        for frame_id, frame_json_resp in sub_job_resp.items():
            children_json_resp[frame_id] = {**children_json_resp[frame_id], **frame_json_resp}

    return children_json_resp


def _classification_annotation_to_json_response(
    annotation: ClassificationAnnotation,
    other_annotations: List[ClassicAnnotation],
) -> Dict[JobName, Dict]:
    # initialize the json response
    json_resp = {
        annotation["job"]: {
            "categories": [],
        }
    }

    # get the child annotations of the current annotation
    # and compute the json response of those child jobs
    child_annotations = _get_child_annotations(annotation, other_annotations)
    json_resp_child_jobs = (
        _compute_children_json_resp(child_annotations, other_annotations)
        if child_annotations
        else {}
    )

    # a frame can have one or multiple categories
    categories = annotation["annotationValue"]["categories"]

    for category in categories:
        category_annotation: Dict = {"name": category}

        # search among the child annotations the ones
        # that have a path (annotationId, category)
        children_json_resp = {}
        for child_ann in child_annotations:
            if [annotation["id"], category] in child_ann["path"] and child_ann[
                "job"
            ] in json_resp_child_jobs:
                children_json_resp[child_ann["job"]] = json_resp_child_jobs[child_ann["job"]]

        if children_json_resp:
            category_annotation["children"] = children_json_resp

        json_resp[annotation["job"]]["categories"].append(category_annotation)

    return json_resp


def _video_classification_annotation_to_json_response(
    annotation: VideoClassificationAnnotation,
    other_annotations: List[VideoAnnotation],
) -> Dict[str, Dict[JobName, Dict]]:
    # get the child annotations of the current annotation
    # and compute the json response of those child jobs
    child_annotations = _get_child_annotations(annotation, other_annotations)
    json_resp_child_jobs = (
        _compute_children_json_resp(child_annotations, other_annotations)
        if child_annotations
        else {}
    )

    json_resp: Dict[str, Dict[JobName, Dict]] = defaultdict(dict)

    for key_ann, key_ann_start, key_ann_end, _, key_ann_frame in _key_annotations_iterator(
        annotation
    ):
        for frame_id in range(key_ann_start, key_ann_end):
            # initialize the frame json response
            json_resp[str(frame_id)][annotation["job"]] = {
                "categories": [],
                "isKeyFrame": frame_id == key_ann_frame,
            }

            # get the frame json response of child jobs
            child_jobs_frame_json_resp = json_resp_child_jobs.get(str(frame_id), {})

            # a frame can have one or multiple categories
            categories = key_ann["annotationValue"]["categories"]
            for category in categories:
                category_annotation: Dict = {"name": category}

                # search among the child annotations the ones
                # that have a path (annotationId, category)
                children_json_resp = {}
                for child_ann in child_annotations:
                    if [annotation["id"], category] in child_ann["path"] and child_ann[
                        "job"
                    ] in child_jobs_frame_json_resp:
                        children_json_resp[child_ann["job"]] = child_jobs_frame_json_resp[
                            child_ann["job"]
                        ]

                if children_json_resp:
                    category_annotation["children"] = children_json_resp

                json_resp[str(frame_id)][annotation["job"]]["categories"].append(
                    category_annotation
                )

    return json_resp


def _video_object_detection_annotation_to_json_response(
    annotation: VideoObjectDetectionAnnotation,
    other_annotations: List[VideoAnnotation],
    json_interface: Dict,
) -> Dict[str, Dict[JobName, Dict]]:
    # get the child annotations of the current annotation
    # and compute the json response of those child jobs
    child_annotations = _get_child_annotations(annotation, other_annotations)
    json_resp_child_jobs = (
        _compute_children_json_resp(child_annotations, other_annotations)
        if child_annotations
        else {}
    )

    json_resp = defaultdict(dict)

    for (
        key_ann,
        key_ann_start,
        key_ann_end,
        next_key_ann,
        key_ann_frame,
    ) in _key_annotations_iterator(annotation):
        for frame_id in range(key_ann_start, key_ann_end):
            # get the frame json response of child jobs
            child_jobs_frame_json_resp = json_resp_child_jobs.get(str(frame_id), {})

            annotation_dict = {
                "children": child_jobs_frame_json_resp,
                "isKeyFrame": frame_id == key_ann_frame,
                "categories": [{"name": annotation["category"]}],
                "mid": annotation["mid"],
                "type": json_interface["jobs"][annotation["job"]]["tools"][0],
            }

            if frame_id == key_ann_frame or next_key_ann is None:
                norm_vertices = key_ann["annotationValue"]["vertices"]
            # between two key frame annotations, an object (point, bbox, polygon) is
            # interpolated in the UI
            else:
                object_inital_state = key_ann["annotationValue"]["vertices"]
                object_final_state = next_key_ann["annotationValue"]["vertices"]
                norm_vertices = _interpolate_object(
                    object_type=json_interface["jobs"][annotation["job"]]["tools"][0],
                    object_initial_state=object_inital_state,
                    initial_state_frame_index=key_ann["frame"],
                    object_final_state=object_final_state,
                    final_state_frame_index=next_key_ann["frame"],
                    at_frame=frame_id,
                )

            if json_interface["jobs"][annotation["job"]]["tools"][0] == "marker":
                annotation_dict["point"] = norm_vertices[0][0][0]

            elif json_interface["jobs"][annotation["job"]]["tools"][0] in {"polygon", "rectangle"}:
                annotation_dict["boundingPoly"] = [{"normalizedVertices": norm_vertices[0][0]}]

            elif json_interface["jobs"][annotation["job"]]["tools"][0] == "semantic":
                annotation_dict["boundingPoly"] = [
                    {"normalizedVertices": norm_vert} for norm_vert in norm_vertices[0]
                ]

            json_resp[str(frame_id)].setdefault(annotation["job"], {}).setdefault(
                "annotations", []
            ).append(annotation_dict)

    return json_resp


def _interpolate_object(
    *,
    object_type: str,
    object_initial_state: List[List[List[Vertice]]],
    initial_state_frame_index: int,
    object_final_state: List[List[List[Vertice]]],
    final_state_frame_index: int,
    at_frame: int,
) -> List[List[List[Vertice]]]:
    """Interpolate an object between two key frames."""
    # if the two frames are consecutive, we do not interpolate
    if at_frame == initial_state_frame_index:
        return object_initial_state

    if at_frame == final_state_frame_index:
        return object_final_state

    if object_type == JobTool.MARKER:
        # for point jobs, we interpolate for each frame
        return [
            [
                [
                    _interpolate_point(
                        previous_point=object_initial_state[0][0][0],
                        next_point=object_final_state[0][0][0],
                        weight=(at_frame - initial_state_frame_index)
                        / (final_state_frame_index - initial_state_frame_index),
                    )
                ]
            ]
        ]

    if object_type == JobTool.RECTANGLE:
        return [
            [
                _interpolate_rectangle(
                    previous_vertices=object_initial_state[0][0],
                    next_vertices=object_final_state[0][0],
                    weight=(at_frame - initial_state_frame_index)
                    / (final_state_frame_index - initial_state_frame_index),
                )
            ]
        ]

    if object_type == JobTool.POLYGON:
        return object_initial_state  # for polygon jobs, we keep the initial state

    if object_type == JobTool.SEMANTIC:
        return object_initial_state  # for semantic jobs, we keep the initial state

    raise NotImplementedError(f"Cannot interpolate object of type {object_type}")


def _interpolate_point(previous_point: Vertice, next_point: Vertice, weight: float) -> Vertice:
    """Interpolate a point."""
    return Vertice(
        x=_interpolate_number(previous_point["x"], next_point["x"], weight),
        y=_interpolate_number(previous_point["y"], next_point["y"], weight),
    )


def _interpolate_rectangle(
    *,
    previous_vertices: List[Vertice],
    next_vertices: List[Vertice],
    weight: float,
) -> List[Vertice]:
    """Interpolate a rectangle.

    It finds a bijection between the rectangles, calculates their properties, and then smoothly
    interpolates the angle, center, length, and width between
    the two rectangles based on the specified weight.
    The interpolated properties are used to reconstruct the vertices of the interpolated rectangle,
    which are then converted back to normalized coordinates.
    """
    permuted_new_vertices = _find_rectangle_vertices_bijection(previous_vertices, next_vertices)

    previous_rectangle_properties = _find_rectangle_properties(previous_vertices)
    next_rectangle_properties = _find_rectangle_properties(permuted_new_vertices)

    interpolated_angle = _interpolate_angle(
        previous_rectangle_properties.angle, next_rectangle_properties.angle, weight
    )
    interpolated_center = _interpolate_point(
        previous_rectangle_properties.center, next_rectangle_properties.center, weight
    )
    interpolated_length = _interpolate_number(
        previous_rectangle_properties.length, next_rectangle_properties.length, weight
    )
    interpolated_width = _interpolate_number(
        previous_rectangle_properties.width, next_rectangle_properties.width, weight
    )

    interpolated_rectangle_properties = _RectangleProperties(
        angle=interpolated_angle,
        center=interpolated_center,
        length=interpolated_length,
        width=interpolated_width,
    )

    interpolated_rectangle = _reconstruct_rectangle_from_properties(
        interpolated_rectangle_properties
    )

    return interpolated_rectangle


def _find_rectangle_vertices_bijection(
    rectangle1: List[Vertice], rectangle2: List[Vertice]
) -> List[Vertice]:
    """Find the bijection between two rectangles that minimizes the cost."""
    permutation_array = [
        [rectangle2[0], rectangle2[1], rectangle2[2], rectangle2[3]],
        [rectangle2[1], rectangle2[2], rectangle2[3], rectangle2[0]],
        [rectangle2[2], rectangle2[3], rectangle2[0], rectangle2[1]],
        [rectangle2[3], rectangle2[0], rectangle2[1], rectangle2[2]],
        [rectangle2[3], rectangle2[2], rectangle2[1], rectangle2[0]],
        [rectangle2[2], rectangle2[1], rectangle2[0], rectangle2[3]],
        [rectangle2[1], rectangle2[0], rectangle2[3], rectangle2[2]],
        [rectangle2[0], rectangle2[3], rectangle2[2], rectangle2[1]],
    ]
    bijection_cost_array = [
        _bijection_cost(permuted_rectangle, rectangle1) for permuted_rectangle in permutation_array
    ]

    optimal_permutation_index = bijection_cost_array.index(min(bijection_cost_array))
    return permutation_array[optimal_permutation_index]


def _bijection_cost(permuted_rectangle: List[Vertice], rectangle: List[Vertice]) -> float:
    """Compute the cost of a bijection between two rectangles."""
    cost = 0
    for i in range(2):
        ab_x = rectangle[i + 1]["x"] - rectangle[i]["x"]
        apermbperm_x = permuted_rectangle[i + 1]["x"] - permuted_rectangle[i]["x"]

        ab_y = rectangle[i + 1]["y"] - rectangle[i]["y"]
        apermbperm_y = permuted_rectangle[i + 1]["y"] - permuted_rectangle[i]["y"]

        ab_norm = math.sqrt(ab_x**2 + ab_y**2)
        apermbnorm = math.sqrt(apermbperm_x**2 + apermbperm_y**2)

        cos = (ab_x * apermbperm_x + ab_y * apermbperm_y) / (ab_norm * apermbnorm)

        cost -= cos

    return cost


@dataclass
class _RectangleProperties:
    """Rectangle properties."""

    angle: float
    center: Vertice
    length: float
    width: float


def _find_rectangle_properties(rectangle: List[Vertice]) -> _RectangleProperties:
    """Find the properties of a rectangle."""
    if len(rectangle) != 4:  # noqa: PLR2004
        raise RuntimeError("Invalid rectangle format")

    angle = _find_rectangle_angle(rectangle)
    center = _find_rectangle_center(rectangle)
    length = _find_rectangle_length(rectangle)
    width = _find_rectangle_width(rectangle)

    return _RectangleProperties(angle=angle, center=center, length=length, width=width)


def _find_rectangle_angle(rectangle: List[Vertice]) -> float:
    """Find the angle of a rectangle."""
    vector_ab = Vertice(
        x=rectangle[2]["x"] - rectangle[1]["x"], y=rectangle[2]["y"] - rectangle[1]["y"]
    )
    return math.atan2(vector_ab["y"], vector_ab["x"])


def _find_rectangle_center(rectangle: List[Vertice]) -> Vertice:
    """Find the center of a rectangle."""
    point_a = rectangle[0]
    point_c = rectangle[2]
    return Vertice(x=(point_a["x"] + point_c["x"]) / 2, y=(point_a["y"] + point_c["y"]) / 2)


def _find_rectangle_length(rectangle: List[Vertice]) -> float:
    """Find the length of a rectangle."""
    return _distance_between_points(rectangle[1], rectangle[2])


def _find_rectangle_width(rectangle: List[Vertice]) -> float:
    """Find the width of a rectangle."""
    return _distance_between_points(rectangle[0], rectangle[1])


def _distance_between_points(point1: Vertice, point2: Vertice) -> float:
    """Compute the distance between two points."""
    return math.sqrt((point1["x"] - point2["x"]) ** 2 + (point1["y"] - point2["y"]) ** 2)


def _interpolate_angle(previous_angle: float, angle: float, weight: float) -> float:
    """Interpolate an angle."""
    difference = min(
        abs(angle - previous_angle),
        2 * math.pi - abs(angle - previous_angle),
    )
    sign = (
        math.copysign(1, angle - previous_angle)
        if abs(angle - previous_angle) < 2 * math.pi - abs(angle - previous_angle)
        else math.copysign(1, 2 * math.pi - abs(angle - previous_angle))
    )
    interpolated_angle = previous_angle + sign * difference * weight
    return interpolated_angle if math.isfinite(interpolated_angle) else 0


def _interpolate_number(previous_number: float, number: float, weight: float) -> float:
    """Interpolate a number."""
    interpolated_number = number * weight + (1 - weight) * previous_number
    return interpolated_number if math.isfinite(interpolated_number) else 0


def _rotate_vector(vector: Vertice, angle: float) -> Vertice:
    """Rotate a vector."""
    return Vertice(
        x=vector["x"] * math.cos(angle) - vector["y"] * math.sin(angle),
        y=vector["x"] * math.sin(angle) + vector["y"] * math.cos(angle),
    )


def _reconstruct_rectangle_from_properties(properties: _RectangleProperties) -> List[Vertice]:
    """Reconstruct a rectangle from its properties."""
    u = Vertice(x=0, y=properties.width)
    v = Vertice(x=properties.length, y=0)

    rotated_u = _rotate_vector(u, properties.angle)
    rotated_v = _rotate_vector(v, properties.angle)

    point_a = Vertice(
        x=properties.center["x"] + (rotated_u["x"] + rotated_v["x"]) / 2,
        y=properties.center["y"] + (rotated_u["y"] + rotated_v["y"]) / 2,
    )
    point_b = Vertice(
        x=properties.center["x"] + (-rotated_u["x"] + rotated_v["x"]) / 2,
        y=properties.center["y"] + (-rotated_u["y"] + rotated_v["y"]) / 2,
    )
    point_c = Vertice(
        x=properties.center["x"] - (rotated_u["x"] + rotated_v["x"]) / 2,
        y=properties.center["y"] - (rotated_u["y"] + rotated_v["y"]) / 2,
    )
    point_d = Vertice(
        x=properties.center["x"] - (-rotated_u["x"] + rotated_v["x"]) / 2,
        y=properties.center["y"] - (-rotated_u["y"] + rotated_v["y"]) / 2,
    )

    return [point_a, point_b, point_c, point_d]



"""Method to import assets from a csv file."""

import csv
from typing import List, Tuple


def get_text_assets_from_csv(from_csv: str, csv_separator: str) -> Tuple[List[str], List[str]]:
    """Get text assets from a csv file."""
    content_array: List[str] = []
    external_id_array: List[str] = []

    with open(from_csv, newline="", encoding="utf-8") as file:
        reader = csv.DictReader(file, delimiter=csv_separator)

        for row in reader:
            content_array.append(row["content"])
            external_id_array.append(row["externalId"])

    return content_array, external_id_array



"""Types for the label-related Kili API gateway functions."""

from dataclasses import dataclass
from typing import Dict, List, Optional

from kili.domain.asset import AssetId
from kili.domain.label import LabelType
from kili.domain.user import UserId


@dataclass
class UpdateLabelData:
    """Data to update a label."""

    is_sent_back_to_queue: Optional[bool]
    json_response: Optional[Dict]
    model_name: Optional[str]
    seconds_to_label: Optional[float]


@dataclass
class AppendLabelData:
    """AppendLabelData data."""

    author_id: Optional[UserId]
    asset_id: AssetId
    client_version: Optional[int]
    json_response: Dict
    seconds_to_label: Optional[float]
    model_name: Optional[str]


@dataclass
class AppendManyLabelsData:
    """AppendManyLabelsData data."""

    labels_data: List[AppendLabelData]
    label_type: LabelType
    overwrite: Optional[bool]


@dataclass
class AppendToLabelsData:
    """AppendToLabelsData graphql input."""

    author_id: UserId
    client_version: Optional[int]
    json_response: Dict
    label_type: LabelType
    seconds_to_label: Optional[float]
    skipped: Optional[bool]



"""CLI's project label subcommand."""

import os
from typing import Optional, Tuple

import click
from typing_extensions import get_args

from kili.domain.project import ProjectId
from kili.entrypoints.cli.common_args import Arguments, Options
from kili.entrypoints.cli.helpers import get_kili_client
from kili.services import label_import
from kili.services.label_import.types import LabelFormat


def type_check_label(key, value):
    """Type check value based on key."""
    if key == "json_response_path" and not (os.path.isfile(value) and value.endswith(".json")):
        return f"{value} is not a valid path to a json file, "

    return ""


@click.command(name="label")
@Options.api_key
@Options.endpoint
@Arguments.files
@Options.project_id
@click.option(
    "--prediction",
    "is_prediction",
    type=bool,
    is_flag=True,
    default=False,
    help=(
        "Tells to import labels as predictions, which means that they will appear "
        "as pre-annotations in the Kili interface"
    ),
)
@click.option(
    "--model-name",
    type=str,
    help="Name of the model that generated predictions, if labels are sent as predictions",
)
@Options.verbose
@click.option(
    "--input-format",
    type=click.Choice(get_args(LabelFormat)),
    help="Format in which the labels are encoded",
    default="kili",
    show_default='"raw" kili format',
)
@click.option(
    "--metadata-file",
    type=str,
    help="File containing format metadata (if relevant to the input format)",
)
@click.option(
    "--target-job",
    type=str,
    help="Job name in the project where to upload the labels (if relevant to the input format)",
)
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help=(
        "when uploading prediction or inference labels, if the `overwrite` flag is activated, it"
        " will overwrite existing labels with the same model name and of the same label type,"
        " on the targeted assets."
    ),
)
# pylint: disable=too-many-arguments
def import_labels(
    api_key: Optional[str],
    endpoint: Optional[str],
    files: Tuple[str, ...],
    project_id: str,
    is_prediction: bool,
    model_name: Optional[str],
    verbose: bool,
    input_format: str,
    metadata_file: Optional[str],
    target_job: Optional[str],
    overwrite: bool,
):
    r"""Import labels or predictions.

    Files can be paths to files or to folders. <br>
    You can provide several paths separated by spaces.
    Label files are JSON files containing labels in the Kili format: the value corresponding to the
    jsonResponse field of a label
    (see [here](https://docs.kili-technology.com/reference/export-classification) for example).<br>
    File's name must be equal to asset's external_id.

    \b
    !!! Examples
        To import default labels:
        ```
        kili project label \\
            dir/labels/ dir/ground-truth/image1.json \\
            --project-id <project_id>
        ```
        To import labels as predictions:
        ```
        kili project label \\
            dir/predictions/ \\
            --project-id <project_id> \\
            --prediction \\
            --model-name YOLO-run-3
        ```
        To import labels as predictions in the Yolo v5 format into a target job:
        ```
        kili project label \\
            dir/predictions/ \\
            --project-id <project_id> \\
            --prediction \\
            --model-name YOLO-v5 \\
            --metadata-file classes.yml \\
            --target-job IMAGE_DETECTION_JOB \\
            --input-format yolo_v5
            --overwrite True
        ```
    """
    if is_prediction and model_name is None:
        raise ValueError(
            "When importing labels as prediction, "
            "you must provide a model name with the --model-name option"
        )

    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint)

    label_import.import_labels_from_files(
        kili,
        list(files or []),
        metadata_file,
        ProjectId(project_id),
        input_format,
        target_job,
        not verbose,
        "INFO" if verbose else "WARNING",
        model_name,
        is_prediction,
        overwrite,
    )



"""Formatters for assets retrieved from Kili API."""

import json
from typing import Dict

from kili.adapters.http_client import HttpClient
from kili.core.helpers import is_url
from kili.domain.types import ListOrTuple


def load_json_from_link(link: str, http_client: HttpClient) -> Dict:
    """Load json from link."""
    if link == "" or not is_url(link):
        return {}

    try:
        return http_client.get(link, timeout=30).json()
    except json.JSONDecodeError:
        return {}


def load_asset_json_fields(asset: Dict, fields: ListOrTuple[str], http_client: HttpClient) -> Dict:
    """Load json fields of an asset."""
    if "jsonMetadata" in fields:
        try:
            asset["jsonMetadata"] = json.loads(asset.get("jsonMetadata", "{}"))
        except json.JSONDecodeError:
            asset["jsonMetadata"] = {}

    if "ocrMetadata" in fields and asset.get("ocrMetadata") is not None:
        asset["ocrMetadata"] = load_json_from_link(asset.get("ocrMetadata", ""), http_client)

    if "labels.jsonResponse" in fields:
        asset_labels = asset.get("labels", [])
        for label in asset_labels:
            try:
                label["jsonResponse"] = json.loads(label["jsonResponse"])
            except json.JSONDecodeError:
                label["jsonResponse"] = {}

    if "latestLabel.jsonResponse" in fields and asset.get("latestLabel") is not None:
        try:
            asset["latestLabel"]["jsonResponse"] = json.loads(asset["latestLabel"]["jsonResponse"])
        except json.JSONDecodeError:
            asset["latestLabel"]["jsonResponse"] = {}

    return asset



"""Helpers for the asset_import."""

import warnings


def is_chat_format(data, required_keys):
    """Checks if llm file data is in chat format."""
    if isinstance(data, dict):
        return False

    if not isinstance(data, list):
        warnings.warn("Json file is not an array.")
        return False

    # Check each item in the array
    for item in data:
        # Ensure each item is a dictionary with the required keys
        if not isinstance(item, dict) or not required_keys.issubset(item.keys()):
            missing_keys = required_keys - set(item.keys())
            raise ValueError(f"Chat item missing keys : {missing_keys}")
    return True


def process_json(data):
    """Processes the llm file data : converts it to Kili format is chat format is present."""
    # Initialize the transformed structure
    transformed_data = {"prompts": [], "type": "markdown", "version": "0.1"}

    # Temporary variables for processing
    current_prompt = None
    completions = []
    models = []  # To store models for determining the last two
    item_ids = []  # To store all item IDs for concatenation
    chat_id = None

    for item in data:
        chat_id = item.get("chat_id", None)
        if item["id"] is not None:
            item_ids.append(item["id"])
        else:
            warnings.warn(f"No id value for chat item {item}.")

        if item["content"] is None:
            raise ValueError("Chat item content cannot be null.")

        # Check if the model is null (indicating a prompt)
        if item["model"] is None:
            # If there's an existing prompt being processed, add it to the prompts list
            if current_prompt is not None:
                transformed_data["prompts"].append(
                    {
                        "completions": completions,
                        "prompt": current_prompt["content"],
                    }
                )
                completions = []  # Reset completions for the next prompt

            # Update the current prompt
            current_prompt = item
        else:
            if item["role"] is None:
                raise ValueError("Chat item role cannot be null.")

            # Add completion to the current prompt
            completions.append(
                {
                    "content": item["content"],
                    "title": item["role"],
                }
            )
            # Collect model for this item
            models.append(item["model"])

    if current_prompt is None:
        raise ValueError(
            "No user prompt found in payload ('model' key set to None) : need at least one."
        )

    # Add the last prompt if it exists
    if current_prompt is not None:
        transformed_data["prompts"].append(
            {
                "completions": completions,
                "prompt": current_prompt["content"],
            }
        )

    chat_item_ids = "_".join(item_ids)

    # Prepare additional_json_metadata
    additional_json_metadata = {
        "chat_id": chat_id,
        "models": "_".join(models[-2:]),  # Join the last two models
        "chat_item_ids": chat_item_ids,  # Concatenate all item IDs
        "text": f"Chat_id: {chat_id}\n\nChat_item_ids: {chat_item_ids}",
    }

    return transformed_data, additional_json_metadata



"""Types used by the conversion service."""

from typing import Callable, Dict, Literal, NamedTuple

ExportType = Literal["latest", "normal"]
SplitOption = Literal["split", "merged"]
LabelFormat = Literal[
    "raw",
    "kili",
    "yolo_v4",
    "yolo_v5",
    "yolo_v7",
    "yolo_v8",
    "coco",
    "pascal_voc",
    "geojson",
    "llm_v1",
]


class JobCategory(NamedTuple):
    """Contains information for a category."""

    category_name: str
    id: int
    job_id: str


CocoAnnotationModifier = Callable[[Dict, Dict, Dict], Dict]



"""Types for the Issue-related Kili API gateway functions."""

from dataclasses import dataclass
from typing import Optional


@dataclass
class IssueToCreateKiliAPIGatewayInput:
    """Data about an issue to create needed in graphql createIssue resolver."""

    label_id: Optional[str]
    object_mid: Optional[str]
    asset_id: str
    text: Optional[str]



"""Client presentation methods for assets."""

import warnings
from typing import (
    TYPE_CHECKING,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Union,
    cast,
    overload,
)

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.domain.asset import AssetExternalId, AssetFilters, AssetId, AssetStatus
from kili.domain.issue import IssueStatus, IssueType
from kili.domain.label import LabelType
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.presentation.client.helpers.common_validators import (
    disable_tqdm_if_as_generator,
)
from kili.use_cases.asset import AssetUseCases
from kili.utils.logcontext import for_all_methods, log_call

from .base import BaseClientMethods

if TYPE_CHECKING:
    import pandas as pd


@for_all_methods(log_call, exclude=["__init__"])
class AssetClientMethods(BaseClientMethods):
    """Methods attached to the Kili client, to run actions on assets."""

    # pylint: disable=too-many-arguments, redefined-builtin, too-many-locals
    @overload
    def assets(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        skip: int = 0,
        fields: ListOrTuple[str] = (
            "content",
            "createdAt",
            "externalId",
            "id",
            "isHoneypot",
            "jsonMetadata",
            "labels.author.id",
            "labels.author.email",
            "labels.createdAt",
            "labels.id",
            "labels.jsonResponse",
            "skipped",
            "status",
        ),
        asset_id_in: Optional[List[str]] = None,
        asset_id_not_in: Optional[List[str]] = None,
        consensus_mark_gt: Optional[float] = None,
        consensus_mark_lt: Optional[float] = None,
        disable_tqdm: Optional[bool] = None,
        external_id_contains: Optional[List[str]] = None,
        first: Optional[int] = None,
        format: Optional[str] = None,
        honeypot_mark_gt: Optional[float] = None,
        honeypot_mark_lt: Optional[float] = None,
        label_author_in: Optional[List[str]] = None,
        label_consensus_mark_gt: Optional[float] = None,
        label_consensus_mark_lt: Optional[float] = None,
        label_created_at: Optional[str] = None,
        label_created_at_gt: Optional[str] = None,
        label_created_at_lt: Optional[str] = None,
        label_honeypot_mark_gt: Optional[float] = None,
        label_honeypot_mark_lt: Optional[float] = None,
        label_type_in: Optional[List[LabelType]] = None,
        metadata_where: Optional[dict] = None,
        skipped: Optional[bool] = None,
        status_in: Optional[List[AssetStatus]] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        label_category_search: Optional[str] = None,
        download_media: bool = False,
        local_media_dir: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        consensus_mark_gte: Optional[float] = None,
        consensus_mark_lte: Optional[float] = None,
        inference_mark_gte: Optional[float] = None,
        inference_mark_lte: Optional[float] = None,
        label_reviewer_in: Optional[List[str]] = None,
        label_consensus_mark_gte: Optional[float] = None,
        label_consensus_mark_lte: Optional[float] = None,
        label_created_at_gte: Optional[str] = None,
        label_created_at_lte: Optional[str] = None,
        label_honeypot_mark_gte: Optional[float] = None,
        label_honeypot_mark_lte: Optional[float] = None,
        issue_type: Optional[IssueType] = None,
        issue_status: Optional[IssueStatus] = None,
        external_id_strictly_in: Optional[List[str]] = None,
        external_id_in: Optional[List[str]] = None,
        label_output_format: Literal["dict", "parsed_label"] = "dict",
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def assets(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        skip: int = 0,
        fields: ListOrTuple[str] = (
            "content",
            "createdAt",
            "externalId",
            "id",
            "isHoneypot",
            "jsonMetadata",
            "labels.author.id",
            "labels.author.email",
            "labels.createdAt",
            "labels.id",
            "labels.jsonResponse",
            "skipped",
            "status",
        ),
        asset_id_in: Optional[List[str]] = None,
        asset_id_not_in: Optional[List[str]] = None,
        consensus_mark_gt: Optional[float] = None,
        consensus_mark_lt: Optional[float] = None,
        disable_tqdm: Optional[bool] = None,
        external_id_contains: Optional[List[str]] = None,
        first: Optional[int] = None,
        format: Optional[str] = None,
        honeypot_mark_gt: Optional[float] = None,
        honeypot_mark_lt: Optional[float] = None,
        label_author_in: Optional[List[str]] = None,
        label_consensus_mark_gt: Optional[float] = None,
        label_consensus_mark_lt: Optional[float] = None,
        label_created_at: Optional[str] = None,
        label_created_at_gt: Optional[str] = None,
        label_created_at_lt: Optional[str] = None,
        label_honeypot_mark_gt: Optional[float] = None,
        label_honeypot_mark_lt: Optional[float] = None,
        label_type_in: Optional[List[LabelType]] = None,
        metadata_where: Optional[dict] = None,
        skipped: Optional[bool] = None,
        status_in: Optional[List[AssetStatus]] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        label_category_search: Optional[str] = None,
        download_media: bool = False,
        local_media_dir: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        consensus_mark_gte: Optional[float] = None,
        consensus_mark_lte: Optional[float] = None,
        inference_mark_gte: Optional[float] = None,
        inference_mark_lte: Optional[float] = None,
        label_reviewer_in: Optional[List[str]] = None,
        label_consensus_mark_gte: Optional[float] = None,
        label_consensus_mark_lte: Optional[float] = None,
        label_created_at_gte: Optional[str] = None,
        label_created_at_lte: Optional[str] = None,
        label_honeypot_mark_gte: Optional[float] = None,
        label_honeypot_mark_lte: Optional[float] = None,
        issue_type: Optional[Literal["QUESTION", "ISSUE"]] = None,
        issue_status: Optional[Literal["OPEN", "SOLVED"]] = None,
        external_id_strictly_in: Optional[List[str]] = None,
        external_id_in: Optional[List[str]] = None,
        label_output_format: Literal["dict", "parsed_label"] = "dict",
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def assets(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        skip: int = 0,
        fields: ListOrTuple[str] = (
            "content",
            "createdAt",
            "externalId",
            "id",
            "isHoneypot",
            "jsonMetadata",
            "labels.author.id",
            "labels.author.email",
            "labels.createdAt",
            "labels.id",
            "labels.jsonResponse",
            "skipped",
            "status",
        ),
        asset_id_in: Optional[List[str]] = None,
        asset_id_not_in: Optional[List[str]] = None,
        consensus_mark_gt: Optional[float] = None,
        consensus_mark_lt: Optional[float] = None,
        disable_tqdm: Optional[bool] = None,
        external_id_contains: Optional[List[str]] = None,
        first: Optional[int] = None,
        format: Optional[str] = None,
        honeypot_mark_gt: Optional[float] = None,
        honeypot_mark_lt: Optional[float] = None,
        label_author_in: Optional[List[str]] = None,
        label_consensus_mark_gt: Optional[float] = None,
        label_consensus_mark_lt: Optional[float] = None,
        label_created_at: Optional[str] = None,
        label_created_at_gt: Optional[str] = None,
        label_created_at_lt: Optional[str] = None,
        label_honeypot_mark_gt: Optional[float] = None,
        label_honeypot_mark_lt: Optional[float] = None,
        label_type_in: Optional[List[LabelType]] = None,
        metadata_where: Optional[dict] = None,
        skipped: Optional[bool] = None,
        status_in: Optional[List[AssetStatus]] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        label_category_search: Optional[str] = None,
        download_media: bool = False,
        local_media_dir: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        consensus_mark_gte: Optional[float] = None,
        consensus_mark_lte: Optional[float] = None,
        inference_mark_gte: Optional[float] = None,
        inference_mark_lte: Optional[float] = None,
        label_reviewer_in: Optional[List[str]] = None,
        label_consensus_mark_gte: Optional[float] = None,
        label_consensus_mark_lte: Optional[float] = None,
        label_created_at_gte: Optional[str] = None,
        label_created_at_lte: Optional[str] = None,
        label_honeypot_mark_gte: Optional[float] = None,
        label_honeypot_mark_lte: Optional[float] = None,
        issue_type: Optional[Literal["QUESTION", "ISSUE"]] = None,
        issue_status: Optional[Literal["OPEN", "SOLVED"]] = None,
        external_id_strictly_in: Optional[List[str]] = None,
        external_id_in: Optional[List[str]] = None,
        label_output_format: Literal["dict", "parsed_label"] = "dict",
        *,
        as_generator: bool = False,
    ) -> Union[Iterable[Dict], "pd.DataFrame"]:
        # pylint: disable=line-too-long
        """Get an asset list, an asset generator or a pandas DataFrame that match a set of constraints.

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset to retrieve.
            asset_id_in: A list of the IDs of the assets to retrieve.
            asset_id_not_in: A list of the IDs of the assets to exclude.
            skip: Number of assets to skip (they are ordered by their date of creation, first to last).
            fields: All the fields to request among the possible fields for the assets.
                    See [the documentation](https://docs.kili-technology.com/reference/graphql-api#asset) for all possible fields.
            first: Maximum number of assets to return.
            consensus_mark_gt: Deprecated. Use `consensus_mark_gte` instead.
            consensus_mark_lt: Deprecated. Use `consensus_mark_lte` instead.
            external_id_contains: Deprecated. Use `external_id_strictly_in` instead.
            metadata_where: Filters by the values of the metadata of the asset.
            honeypot_mark_gt: Deprecated. Use `honeypot_mark_gte` instead.
            honeypot_mark_lt: Deprecated. Use `honeypot_mark_lte` instead.
            status_in: Returned assets should have a status that belongs to that list, if given.
                Possible choices: `TODO`, `ONGOING`, `LABELED`, `TO_REVIEW` or `REVIEWED`.
            label_type_in: Returned assets should have a label whose type belongs to that list, if given.
            label_author_in: Returned assets should have a label whose author belongs to that list, if given. An author can be designated by the first name, the last name, or the first name + last name.
            label_consensus_mark_gt: Deprecated. Use `label_consensus_mark_gte` instead.
            label_consensus_mark_lt: Deprecated. Use `label_consensus_mark_lte` instead.
            label_created_at: Returned assets should have a label whose creation date is equal to this date.
            label_created_at_gt: Deprecated. Use `label_created_at_gte` instead.
            label_created_at_lt: Deprecated. Use `label_created_at_lte` instead.
            label_honeypot_mark_gt: Deprecated. Use `label_honeypot_mark_gte` instead.
            label_honeypot_mark_lt: Deprecated. Use `label_honeypot_mark_lte` instead.
            skipped: Returned assets should be skipped
            updated_at_gte: Returned assets should have a label whose update date is greater or equal to this date.
            updated_at_lte: Returned assets should have a label whose update date is lower or equal to this date.
            format: If equal to 'pandas', returns a pandas DataFrame
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the assets is returned.
            label_category_search: Returned assets should have a label that follows this category search query.
            download_media: Tell is the media have to be downloaded or not.
            local_media_dir: Directory where the media are downloaded if `download_media` is True.
            created_at_gte: Returned assets should have their import date greater or equal to this date.
            created_at_lte: Returned assets should have their import date lower or equal to this date.
            honeypot_mark_lte: Maximum amount of honeypot for the asset.
            honeypot_mark_gte: Minimum amount of honeypot for the asset.
            consensus_mark_lte: Maximum amount of consensus for the asset.
            consensus_mark_gte: Minimum amount of consensus for the asset.
            inference_mark_gte: Minimum amount of human/model IoU for the asset.
            inference_mark_lte: Maximum amount of human/model IoU for the asset.
            label_reviewer_in: Returned assets should have a label whose reviewer belongs to that list, if given.
            label_consensus_mark_gte: Returned assets should have a label whose consensus is greater or equal to this number.
            label_consensus_mark_lte: Returned assets should have a label whose consensus is lower or equal to this number.
            label_created_at_lte: Returned assets should have a label whose creation date is lower or equal to this date.
            label_created_at_gte: Returned assets should have a label whose creation date is greater or equal to this date.
            label_honeypot_mark_gte: Returned assets should have a label whose honeypot is greater or equal to this number.
            label_honeypot_mark_lte: Returned assets should have a label whose honeypot is lower or equal to this number.
            issue_type: Returned assets should have issues of type `QUESTION` or `ISSUE`.
            issue_status: Returned assets should have issues of status `OPEN` or `SOLVED`.
            external_id_strictly_in: Returned assets should have external ids that match exactly the ones in the list.
            external_id_in: Returned assets should have external ids that partially match the ones in the list.
                For example, with `external_id_in=['abc']`, any asset with an external id containing `'abc'` will be returned.
            label_output_format: If `parsed_label`, the labels in the assets will be parsed. More information on parsed labels in the [documentation](https://python-sdk-docs.kili-technology.com/latest/sdk/tutorials/label_parsing/).

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        !!! info "Filtering by label properties"
            When the assets are filtered by label properties using any of `label_*` filter arguments, as soon as **one**
            label matches **all** the label property criteria, the asset is kept and returned by the method. If any of the
            `labels.*` or `latestLabel.*` subfields are queried, **all** the labels of the kept assets are returned together
            with the assets (and not only the ones matching the criteria)

        Returns:
            An asset list, an asset generator or a pandas DataFrame that match a set of constraints.

        Example:
            ```
            # returns the assets list of the project
            >>> kili.assets(project_id)
            >>> kili.assets(project_id, asset_id=asset_id)
            # returns a generator of the project assets
            >>> kili.assets(project_id, as_generator=True)
            ```

        !!! example "How to filter based on Metadata"
            - `metadata_where = {key1: "value1"}` to filter on assets whose metadata
                have key "key1" with value "value1"
            - `metadata_where = {key1: ["value1", "value2"]}` to filter on assets whose metadata
                have key "key1" with value "value1" or value "value2
            - `metadata_where = {key2: [2, 10]}` to filter on assets whose metadata
                have key "key2" with a value between 2 and 10.

        !!! example "How to filter based on label categories"
            The search query is composed of logical expressions following this format:

                [job_name].[category_name].count [comparaison_operator] [value]
            where:

            - `[job_name]` is the name of the job in the interface
            - `[category_name]` is the name of the category in the interface for this job
            - `[comparaison_operator]` can be one of: [`==`, `>=`, `<=`, `<`, `>`]
            - `[value]` is an integer that represents the count of such objects of the given category in the label

            These operations can be separated by OR and AND operators:
                ```python
                label_category_search = `JOB_CLASSIF.CATEGORY_A.count > 0`
                label_category_search = `JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0`
                label_category_search = `(JOB_CLASSIF.CATEGORY_A.count == 1 OR JOB_NER.CATEGORY_B.count > 0) AND JOB_BBOX.CATEGORY_C.count > 10`
                ```
        """
        if format == "pandas" and as_generator:
            raise ValueError(
                'Argument values as_generator==True and format=="pandas" are not compatible.'
            )

        if external_id_contains is not None:
            warnings.warn(
                "external_id_contains is deprecated, use external_id_strictly_in instead",
                DeprecationWarning,
                stacklevel=1,
            )

        for arg_name, arg_value in zip(
            (
                "consensus_mark_gt",
                "consensus_mark_lt",
                "honeypot_mark_gt",
                "honeypot_mark_lt",
                "label_consensus_mark_gt",
                "label_consensus_mark_lt",
                "label_created_at_gt",
                "label_created_at_lt",
                "label_honeypot_mark_gt",
                "label_honeypot_mark_lt",
            ),
            (
                consensus_mark_gt,
                consensus_mark_lt,
                honeypot_mark_gt,
                honeypot_mark_lt,
                label_consensus_mark_gt,
                label_consensus_mark_lt,
                label_created_at_gt,
                label_created_at_lt,
                label_honeypot_mark_gt,
                label_honeypot_mark_lt,
            ),
        ):
            if arg_value:
                warnings.warn(
                    f"'{arg_name}' is deprecated, please use"
                    f" '{arg_name.replace('_gt', '_gte').replace('_lt', '_lte')}' instead.",
                    DeprecationWarning,
                    stacklevel=1,
                )

        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)

        asset_use_cases = AssetUseCases(self.kili_api_gateway)
        filters = AssetFilters(
            project_id=ProjectId(project_id),
            asset_id=AssetId(asset_id) if asset_id else None,
            asset_id_in=cast(List[AssetId], asset_id_in) if asset_id_in else None,
            asset_id_not_in=cast(List[AssetId], asset_id_not_in) if asset_id_not_in else None,
            consensus_mark_gte=consensus_mark_gt or consensus_mark_gte,
            consensus_mark_lte=consensus_mark_lt or consensus_mark_lte,
            external_id_strictly_in=(
                cast(List[AssetExternalId], external_id_strictly_in or external_id_contains)
                if external_id_strictly_in or external_id_contains
                else None
            ),
            external_id_in=cast(List[AssetExternalId], external_id_in) if external_id_in else None,
            honeypot_mark_gte=honeypot_mark_gt or honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lt or honeypot_mark_lte,
            inference_mark_gte=inference_mark_gte,
            inference_mark_lte=inference_mark_lte,
            label_author_in=label_author_in,
            label_consensus_mark_gte=label_consensus_mark_gt or label_consensus_mark_gte,
            label_consensus_mark_lte=label_consensus_mark_lt or label_consensus_mark_lte,
            label_created_at=label_created_at,
            label_created_at_gte=label_created_at_gt or label_created_at_gte,
            label_created_at_lte=label_created_at_lt or label_created_at_lte,
            label_honeypot_mark_gte=label_honeypot_mark_gt or label_honeypot_mark_gte,
            label_honeypot_mark_lte=label_honeypot_mark_lt or label_honeypot_mark_lte,
            label_type_in=label_type_in,
            metadata_where=metadata_where,
            skipped=skipped,
            status_in=status_in,
            updated_at_gte=updated_at_gte,
            updated_at_lte=updated_at_lte,
            label_category_search=label_category_search,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            label_reviewer_in=label_reviewer_in,
            issue_status=issue_status,
            issue_type=issue_type,
        )
        assets_gen = asset_use_cases.list_assets(
            filters,
            fields,
            download_media=download_media,
            local_media_dir=local_media_dir,
            label_output_format=label_output_format,
            options=QueryOptions(disable_tqdm=disable_tqdm, first=first, skip=skip),
        )

        if format == "pandas":
            import pandas as pd  # pylint: disable=import-outside-toplevel

            return pd.DataFrame(list(assets_gen))

        if as_generator:
            return assets_gen
        return list(assets_gen)

    # pylint: disable=too-many-arguments,too-many-locals
    @typechecked
    def count_assets(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_id_in: Optional[List[str]] = None,
        asset_id_not_in: Optional[List[str]] = None,
        external_id_contains: Optional[List[str]] = None,
        metadata_where: Optional[dict] = None,
        status_in: Optional[List[AssetStatus]] = None,
        consensus_mark_gt: Optional[float] = None,
        consensus_mark_lt: Optional[float] = None,
        honeypot_mark_gt: Optional[float] = None,
        honeypot_mark_lt: Optional[float] = None,
        label_type_in: Optional[List[LabelType]] = None,
        label_author_in: Optional[List[str]] = None,
        label_consensus_mark_gt: Optional[float] = None,
        label_consensus_mark_lt: Optional[float] = None,
        label_created_at: Optional[str] = None,
        label_created_at_gt: Optional[str] = None,
        label_created_at_lt: Optional[str] = None,
        label_honeypot_mark_gt: Optional[float] = None,
        label_honeypot_mark_lt: Optional[float] = None,
        skipped: Optional[bool] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        label_category_search: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        consensus_mark_gte: Optional[float] = None,
        consensus_mark_lte: Optional[float] = None,
        inference_mark_gte: Optional[float] = None,
        inference_mark_lte: Optional[float] = None,
        label_reviewer_in: Optional[List[str]] = None,
        label_consensus_mark_gte: Optional[float] = None,
        label_consensus_mark_lte: Optional[float] = None,
        label_created_at_gte: Optional[str] = None,
        label_created_at_lte: Optional[str] = None,
        label_honeypot_mark_gte: Optional[float] = None,
        label_honeypot_mark_lte: Optional[float] = None,
        issue_type: Optional[IssueType] = None,
        issue_status: Optional[IssueStatus] = None,
        external_id_strictly_in: Optional[List[str]] = None,
        external_id_in: Optional[List[str]] = None,
    ) -> int:
        # pylint: disable=line-too-long
        """Count and return the number of assets with the given constraints.

        Parameters beginning with 'label_' apply to labels, others apply to assets.

        Args:
            project_id: Identifier of the project
            asset_id: The unique id of the asset to retrieve.
            asset_id_in: A list of the ids of the assets to retrieve.
            asset_id_not_in: A list of the ids of the assets to exclude.
            external_id_contains: Deprecated. Use `external_id_strictly_in` instead.
            metadata_where: Filters by the values of the metadata of the asset.
            status_in: Returned assets should have a status that belongs to that list, if given. Possible choices: `TODO`, `ONGOING`, `LABELED`, `TO_REVIEW` or `REVIEWED`.
            consensus_mark_gt: Deprecated. Use `consensus_mark_gte` instead.
            consensus_mark_lt: Deprecated. Use `consensus_mark_lte` instead.
            honeypot_mark_gt: Deprecated. Use `honeypot_mark_gte` instead.
            honeypot_mark_lt: Deprecated. Use `honeypot_mark_lte` instead.
            label_type_in: Returned assets should have a label whose type belongs to that list, if given.
            label_author_in: Returned assets should have a label whose author belongs to that list, if given. An author can be designated by the first name, the last name, or the first name + last name.
            label_consensus_mark_gt: Deprecated. Use `label_consensus_mark_gte` instead.
            label_consensus_mark_lt: Deprecated. Use `label_consensus_mark_lte` instead.
            label_created_at: Returned assets should have a label whose creation date is equal to this date.
            label_created_at_gt: Deprecated. Use `label_created_at_gte` instead.
            label_created_at_lt: Deprecated. Use `label_created_at_lte` instead.
            label_honeypot_mark_gt: Deprecated. Use `label_honeypot_mark_gte` instead.
            label_honeypot_mark_lt: Deprecated. Use `label_honeypot_mark_lte` instead.
            skipped: Returned assets should be skipped.
            updated_at_gte: Returned assets should have a label whose update date is greated or equal to this date.
            updated_at_lte: Returned assets should have a label whose update date is lower or equal to this date.
            label_category_search: Returned assets should have a label that follows this category search query.
            created_at_gte: Returned assets should have their import date greater or equal to this date.
            created_at_lte: Returned assets should have their import date lower or equal to this date.
            honeypot_mark_lte: Maximum amount of honeypot for the asset.
            honeypot_mark_gte: Minimum amount of honeypot for the asset.
            consensus_mark_lte: Maximum amount of consensus for the asset.
            consensus_mark_gte: Minimum amount of consensus for the asset.
            inference_mark_gte: Minimum amount of human/model IoU for the asset.
            inference_mark_lte: Maximum amount of human/model IoU for the asset.
            label_reviewer_in: Returned assets should have a label whose reviewer belongs to that list, if given.
            label_consensus_mark_gte: Returned assets should have a label whose consensus is greater or equal to this number.
            label_consensus_mark_lte: Returned assets should have a label whose consensus is lower or equal to this number.
            label_created_at_lte: Returned assets should have a label whose creation date is lower or equal to this date.
            label_created_at_gte: Returned assets should have a label whose creation date is greater or equal to this date.
            label_honeypot_mark_gte: Returned assets should have a label whose honeypot is greater or equal to this number.
            label_honeypot_mark_lte: Returned assets should have a label whose honeypot is lower or equal to this number.
            issue_type: Returned assets should have issues of type `QUESTION` or `ISSUE`.
            issue_status: Returned assets should have issues of status `OPEN` or `SOLVED`.
            external_id_strictly_in: Returned assets should have external ids that match exactly the ones in the list.
            external_id_in: Returned assets should have external ids that partially match the ones in the list.
                For example, with `external_id_in=['abc']`, any asset with an external id containing `'abc'` will be returned.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            The number of assets that match the given constraints.

        Examples:
            >>> kili.count_assets(project_id=project_id)
            250
            >>> kili.count_assets(asset_id=asset_id)
            1

        !!! example "How to filter based on Metadata"
            - `metadata_where = {key1: "value1"}` to filter on assets whose metadata
                have key "key1" with value "value1"
            - `metadata_where = {key1: ["value1", "value2"]}` to filter on assets whose metadata
                have key "key1" with value "value1" or value "value2
            - `metadata_where = {key2: [2, 10]}` to filter on assets whose metadata
                have key "key2" with a value between 2 and 10.
        """
        if external_id_contains is not None:
            warnings.warn(
                "external_id_contains is deprecated, use external_id_strictly_in instead",
                DeprecationWarning,
                stacklevel=1,
            )

        for arg_name, arg_value in zip(
            (
                "consensus_mark_gt",
                "consensus_mark_lt",
                "honeypot_mark_gt",
                "honeypot_mark_lt",
                "label_consensus_mark_gt",
                "label_consensus_mark_lt",
                "label_created_at_gt",
                "label_created_at_lt",
                "label_honeypot_mark_gt",
                "label_honeypot_mark_lt",
            ),
            (
                consensus_mark_gt,
                consensus_mark_lt,
                honeypot_mark_gt,
                honeypot_mark_lt,
                label_consensus_mark_gt,
                label_consensus_mark_lt,
                label_created_at_gt,
                label_created_at_lt,
                label_honeypot_mark_gt,
                label_honeypot_mark_lt,
            ),
        ):
            if arg_value:
                warnings.warn(
                    f"'{arg_name}' is deprecated, please use"
                    f" '{arg_name.replace('_gt', '_gte').replace('_lt', '_lte')}' instead.",
                    DeprecationWarning,
                    stacklevel=1,
                )

        filters = AssetFilters(
            project_id=ProjectId(project_id),
            asset_id=AssetId(asset_id) if asset_id else None,
            asset_id_in=cast(List[AssetId], asset_id_in) if asset_id_in else None,
            asset_id_not_in=cast(List[AssetId], asset_id_not_in) if asset_id_not_in else None,
            consensus_mark_gte=consensus_mark_gt or consensus_mark_gte,
            consensus_mark_lte=consensus_mark_lt or consensus_mark_lte,
            external_id_strictly_in=(
                cast(List[AssetExternalId], external_id_strictly_in or external_id_contains)
                if external_id_strictly_in or external_id_contains
                else None
            ),
            external_id_in=cast(List[AssetExternalId], external_id_in) if external_id_in else None,
            honeypot_mark_gte=honeypot_mark_gt or honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lt or honeypot_mark_lte,
            inference_mark_gte=inference_mark_gte,
            inference_mark_lte=inference_mark_lte,
            label_author_in=label_author_in,
            label_reviewer_in=label_reviewer_in,
            label_consensus_mark_gte=label_consensus_mark_gt or label_consensus_mark_gte,
            label_consensus_mark_lte=label_consensus_mark_lt or label_consensus_mark_lte,
            label_created_at=label_created_at,
            label_created_at_gte=label_created_at_gt or label_created_at_gte,
            label_created_at_lte=label_created_at_lt or label_created_at_lte,
            label_honeypot_mark_gte=label_honeypot_mark_gt or label_honeypot_mark_gte,
            label_honeypot_mark_lte=label_honeypot_mark_lt or label_honeypot_mark_lte,
            label_type_in=label_type_in,
            metadata_where=metadata_where,
            skipped=skipped,
            status_in=status_in,
            updated_at_gte=updated_at_gte,
            updated_at_lte=updated_at_lte,
            label_category_search=label_category_search,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            issue_status=issue_status,
            issue_type=issue_type,
        )
        asset_use_cases = AssetUseCases(self.kili_api_gateway)
        return asset_use_cases.count_assets(filters)



"""Kili API Gateway module for interacting with Kili."""

from kili.adapters.http_client import HttpClient
from kili.adapters.kili_api_gateway.api_key.operations_mixin import ApiKeyOperationMixin
from kili.adapters.kili_api_gateway.asset.operations_mixin import AssetOperationMixin
from kili.adapters.kili_api_gateway.cloud_storage import CloudStorageOperationMixin
from kili.adapters.kili_api_gateway.issue import IssueOperationMixin
from kili.adapters.kili_api_gateway.label.operations_mixin import LabelOperationMixin
from kili.adapters.kili_api_gateway.notification.operations_mixin import (
    NotificationOperationMixin,
)
from kili.adapters.kili_api_gateway.organization.operations_mixin import (
    OrganizationOperationMixin,
)
from kili.adapters.kili_api_gateway.project.operations_mixin import ProjectOperationMixin
from kili.adapters.kili_api_gateway.tag import TagOperationMixin
from kili.adapters.kili_api_gateway.user.operation_mixin import UserOperationMixin
from kili.core.graphql.graphql_client import GraphQLClient


class KiliAPIGateway(
    ApiKeyOperationMixin,
    AssetOperationMixin,
    CloudStorageOperationMixin,
    IssueOperationMixin,
    LabelOperationMixin,
    NotificationOperationMixin,
    OrganizationOperationMixin,
    ProjectOperationMixin,
    TagOperationMixin,
    UserOperationMixin,
):
    """GraphQL gateway to communicate with Kili backend."""

    def __init__(self, graphql_client: GraphQLClient, http_client: HttpClient) -> None:
        """Initialize the Kili API Gateway."""
        self.graphql_client = graphql_client
        self.http_client = http_client



"""Label gateway common."""
from kili.adapters.kili_api_gateway.helpers.queries import fragment_builder
from kili.adapters.kili_api_gateway.label.operations import (
    get_annotations_partial_query,
)


def get_annotation_fragment():
    """Generates a fragment to get all annotations and their values."""
    return get_annotations_partial_query(
        annotation_fragment=fragment_builder(("__typename", "id", "job", "path", "labelId")),
        classification_annotation_fragment=fragment_builder(("annotationValue.categories",)),
        ranking_annotation_fragment=fragment_builder(
            (
                "annotationValue.orders.elements",
                "annotationValue.orders.rank",
            )
        ),
        transcription_annotation_fragment=fragment_builder(("annotationValue.text",)),
        video_annotation_fragment=fragment_builder(
            (
                "frames.start",
                "frames.end",
                "keyAnnotations.id",
                "keyAnnotations.frame",
            )
        ),
        video_classification_annotation_fragment=fragment_builder(
            ("keyAnnotations.annotationValue.categories",)
        ),
        video_object_detection_annotation_fragment=fragment_builder(
            (
                "keyAnnotations.annotationValue.vertices.x",
                "keyAnnotations.annotationValue.vertices.y",
                "name",
                "mid",
                "category",
            )
        ),
        video_transcription_annotation_fragment=fragment_builder(
            ("keyAnnotations.annotationValue.text",)
        ),
    )



"""Types for Issue-related use cases."""

from dataclasses import dataclass
from typing import Optional

from kili.domain.label import LabelId


@dataclass
class IssueToCreateUseCaseInput:
    """Data about one Issue to create."""

    label_id: LabelId
    object_mid: Optional[str] = None
    text: Optional[str] = None
